"""Project Issue model class."""

from typing import Dict, Any, List, Optional, Union
from dataclasses import dataclass
from datetime import datetime
from ..client import GraphQLClient
from ..queries import load_query
from .comment import Comment
from .events import Event
from .field import Field
from .label import Label
from .user import User


@dataclass
class Issue:
    """Represents an issue or pull request in a GitHub Project (V2)."""

    id: str = ""
    number: int = 0
    title: str = ""
    url: str = ""
    state: str = ""
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    field_values: List[Dict[str, Any]] = None
    author_login: str = ""
    body: Optional[str] = None
    client: Optional[GraphQLClient] = None
    owner: Optional[str] = None
    repo: Optional[str] = None
    
    def __post_init__(self):
        """Initialize default values for list fields."""
        if self.field_values is None:
            self.field_values = []
        if self.created_at is None:
            self.created_at = datetime.now()
        if self.updated_at is None:
            self.updated_at = datetime.now()

    @classmethod
    def from_response(
        cls,
        data: Dict[str, Any],
        client: Optional[GraphQLClient] = None,
        owner: Optional[str] = None,
        repo: Optional[str] = None,
    ) -> "Issue":
        """Create an Issue instance from API response data.

        Args:
            data (Dict[str, Any]): Raw issue data from API response
            client (Optional[GraphQLClient]): GraphQL client for API calls
            owner (Optional[str]): Owner of the repository
            repo (Optional[str]): Repository name

        Returns:
            Issue: New Issue instance
        """
        # Handle both issue-specific and project item responses
        if "content" in data:  # Project item response
            content = data["content"]
            return cls(
                id=content["id"],
                number=content["number"],
                title=content["title"],
                url=content["url"],
                state=content["state"],
                created_at=datetime.fromisoformat(
                    content["createdAt"].replace("Z", "+00:00")
                ),
                updated_at=datetime.fromisoformat(
                    content["updatedAt"].replace("Z", "+00:00")
                ),
                author_login=content.get("author", {}).get("login", ""),
                body=None,  # Not included in project responses
                field_values=data.get("fieldValues", {"nodes": []})["nodes"],
                client=client,
                owner=owner,
                repo=repo,
            )
        else:  # Issue-specific response
            # For search results the data is in ["node"], directly in data, or under repository.issue
            if "node" in data:
                issue_data = data["node"]
            elif "repository" in data:
                issue_data = data["repository"]["issue"]
            else:
                issue_data = data

            return cls(
                id=issue_data["id"],
                number=issue_data["number"],
                title=issue_data["title"],
                url=issue_data["url"],
                state=issue_data["state"],
                created_at=datetime.fromisoformat(
                    issue_data["createdAt"].replace("Z", "+00:00")
                ),
                updated_at=datetime.fromisoformat(
                    issue_data["updatedAt"].replace("Z", "+00:00")
                ),
                author_login=issue_data["author"]["login"],
                body=issue_data.get("body"),
                field_values=[],  # Not included in issue responses
                client=client,
                owner=owner,
                repo=repo,
            )

    def get_comments(self, first: int = 20) -> Dict[str, List[Comment]]:
        """Get comments on this issue.

        Args:
            first (int, optional): Number of comments to fetch per page. Defaults to 20.

        Returns:
            dict: Issue comments as Comment objects, in "nodes" key

        Raises:
            ValueError: If client, owner or repo are not set
            GraphQLError: On GraphQL operation failures
        """
        if not all([self.client, self.owner, self.repo]):
            raise ValueError("client, owner and repo must be set to fetch comments")

        query = load_query("get_issue_comments")
        variables = {
            "owner": self.owner,
            "repo": self.repo,
            "number": self.number,
            "first": first,
        }

        response = self.client.execute(query, variables)
        comments_data = response["repository"]["issue"]["comments"]
        comments_data["nodes"] = [
            Comment.from_response(node) for node in comments_data["nodes"]
        ]
        return comments_data

    def get_timeline(self, first: int = 20) -> List[Event]:
        """Get timeline events on this issue.

        Args:
            first (int, optional): Number of events to fetch per page. Defaults to 20.

        Returns:
            List[Event]: Issue timeline events as Event objects

        Raises:
            ValueError: If client, owner or repo are not set
            GraphQLError: On GraphQL operation failures
        """
        if not all([self.client, self.owner, self.repo]):
            raise ValueError("client, owner and repo must be set to fetch timeline events")

        query = load_query("get_issue_timeline")
        variables = {
            "owner": self.owner,
            "repo": self.repo,
            "number": self.number,
            "first": first,
        }

        response = self.client.execute(query, variables)
        timeline_data = response["repository"]["issue"]["timelineItems"]
        nodes = [
            Event.create_from_response(node) for node in timeline_data["nodes"] if Event.create_from_response(node)
        ]
        return nodes
        
    def get_labels(self, first: int = 20, after: Optional[str] = None) -> List[Label]:
        """Get labels on this issue.

        Args:
            first (int, optional): Number of labels to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.

        Returns:
            List[Label]: List of Label objects for the issue

        Raises:
            ValueError: If client, owner or repo are not set
            GraphQLError: On GraphQL operation failures
        """
        if not all([self.client, self.owner, self.repo]):
            raise ValueError("client, owner and repo must be set to fetch labels")

        query = load_query("get_issue_labels")
        variables = {
            "owner": self.owner,
            "repo": self.repo,
            "number": self.number,
            "first": first,
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        labels_data = response["repository"]["issue"]["labels"]
        
        # Convert each node to a Label object and return the list directly
        return [
            Label(
                id=label["id"],
                name=label["name"],
                color=label["color"],
                description=label.get("description"),
                url=label.get("url"),
                created_at=Label.from_response({"repository": {"label": label}}).created_at,
                updated_at=Label.from_response({"repository": {"label": label}}).updated_at,
                is_default=label.get("isDefault", False),
                client=self.client
            ) for label in labels_data["nodes"]
        ]
        
    def get_fields(self, first: int = 20) -> List[Field]:
        """Get field values for this issue as Field objects.

        This method converts the raw field values from the API into Field instances.
        If field_values are already loaded, it will use those values.
        Otherwise, it will raise a ValueError as field values can only be loaded from a project.

        Args:
            first (int, optional): Number of fields to fetch per page. Defaults to 20.

        Returns:
            List[Field]: List of Field objects representing the issue's field values

        Raises:
            ValueError: If no field values are available
        """
        if not self.field_values:
            raise ValueError("No field values available - field values can only be loaded from a project")

        # Convert field values to Field instances
        fields = []
        for field_value in self.field_values:
            # Create a field data dictionary with the necessary fields
            field_data = {
                "id": field_value.get("field", {}).get("id", ""),
                "name": field_value.get("field", {}).get("name", ""),
                "type": field_value.get("__typename", ""),
                "dataType": field_value.get("field", {}).get("dataType") or field_value.get("__typename", "").replace("ProjectV2ItemField", "").replace("Value", "").upper()
            }
            
            # Fix for SingleSelect data type format
            if field_data["dataType"] == "SINGLESELECT":
                field_data["dataType"] = "SINGLE_SELECT"
            
            # Add field value based on the type
            if "text" in field_value:
                field_data["value"] = field_value["text"]
            elif "date" in field_value:
                field_data["value"] = field_value["date"]
            elif "name" in field_value:
                field_data["value"] = field_value["name"]
            # Handle user, repository, and milestone field types
            elif field_value.get("__typename") in [
                Field.USER,
                Field.REPOSITORY, 
                Field.MILESTONE
            ]:
                if field_value.get("__typename") == Field.USER:
                    # Handle users field (plural) - create User instances
                    # Users field is now a UserConnection with nodes structure
                    users_connection = field_value.get("users", {})
                    if isinstance(users_connection, dict):
                        users_data = users_connection.get("nodes", [])
                    else:
                        users_data = users_connection

                    if users_data:
                        field_data["value"] = [
                            User(
                                id="",  # Not available in the field response
                                login=user_data.get("login", ""),
                                name=None,
                                email=None,
                                bio=None,
                                company=None,
                                location=None,
                                website_url=None,
                                url="",  # Not available in the field response
                                avatar_url="",  # Not available in the field response
                            )
                            for user_data in users_data
                        ]
                    else:
                        field_data["value"] = []
                else:
                    # Handle repository and milestone fields (unchanged)
                    user_or_repo = field_value.get("repository") or field_value.get("milestone")
                    if user_or_repo:
                        field_data["value"] = user_or_repo.get("name") or user_or_repo.get("title")
                
                # We no longer fetch options for SingleSelect fields
            
            fields.append(Field.from_response(field_data))
        
        return fields
        
    def get(
        self, 
        *,
        client: GraphQLClient, 
        repository: str,
        number: int,
        org: Optional[str] = None,
        username: Optional[str] = None
    ) -> "Issue":
        """Get an issue's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            repository (str): Repository name
            number (int): Issue number
            org (Optional[str]): Organization name (if issue is in an org's repo)
            username (Optional[str]): Username (if issue is in a user's repo)
            
        Returns:
            Issue: Self with populated properties
            
        Raises:
            ValueError: If neither org nor username is provided
            GraphQLError: On GraphQL operation failures
        """
        if not org and not username:
            raise ValueError("Either org or username must be provided")
            
        owner = org if org else username
        query = load_query("get_issue")
        variables = {
            "owner": owner,
            "repo": repository,
            "number": number
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create an issue instance with all properties
        issue = Issue.from_response(response, client, owner, repository)
        
        # Update this instance's attributes with the fetched data
        self.id = issue.id
        self.number = issue.number
        self.title = issue.title
        self.url = issue.url
        self.state = issue.state
        self.created_at = issue.created_at
        self.updated_at = issue.updated_at
        self.author_login = issue.author_login
        self.body = issue.body
        self.field_values = issue.field_values
        self.client = client
        self.owner = owner
        self.repo = repository
        
        return self



"""GitHub Projects V2 API client implementation."""

from .client import GraphQLClient
from .queries import env, load_query
from .models.project import Project
from .models.user import User
from .models.label import Label
from .models.milestone import Milestone
from .models.organization import Organization
from .models.repository import Repository
from typing import Any, Union, List, TYPE_CHECKING
from .clients import _paginate_query, SearchType

if TYPE_CHECKING:
    from .models.issue import Issue
    from .models.field import Field


class ProjectsV2Client:
    """Client for interacting with GitHub's ProjectsV2 API."""

    def __init__(self, client_or_token: Union[GraphQLClient, str], api_url: str = "https://api.github.com/graphql"):
        """Initialize the ProjectsV2 client.

        Args:
            client_or_token (Union[GraphQLClient, str]): Either an authenticated GraphQL client instance 
                                                       or a GitHub Personal Access Token string
            api_url (str, optional): GitHub GraphQL API endpoint. Only used when client_or_token is a token string.
                                   Defaults to "https://api.github.com/graphql".
        
        Raises:
            ValueError: If client_or_token is None or invalid
        """
        if client_or_token is None:
            raise ValueError(
                "client_or_token cannot be None. Please provide either a GitHub Personal Access Token "
                "string or a GraphQLClient instance. If using an environment variable, ensure it is set."
            )
        
        if isinstance(client_or_token, str):
            if not client_or_token.strip():
                raise ValueError(
                    "Token cannot be empty. Please provide a valid GitHub Personal Access Token."
                )
            # Create GraphQLClient internally when token string is provided
            self.client = GraphQLClient(client_or_token, api_url)
        else:
            # Use provided GraphQLClient instance
            self.client = client_or_token

    def get_project(
        self, project_number: int, owner: str, owner_type: str = "user"
    ) -> Project:
        """Get a ProjectV2 by number.

        Args:
            project_number (int): The number of the project
            owner (str): The login of the project owner
            owner_type (str, optional): Type of owner - "user" or "org".
                                      Defaults to "user".

        Returns:
            Project: Project object containing the API data and methods

        Raises:
            ValueError: If owner_type is not "user" or "org"
        """
        if owner_type not in ["user", "org"]:
            raise ValueError('owner_type must be either "user" or "org"')

        # Render the query template with the owner_type context
        tmpl_name = "get_project.graphql"
        query = env.get_template(tmpl_name).render(owner_type=owner_type)

        variables = {"owner": owner, "number": project_number}
        response = self.client.execute(query, variables)

        # Transform response to have consistent structure
        formatted_response = {}
        if owner_type == "user" and response.get("user", {}).get("projectV2"):
            formatted_response = {"user": response["user"]}
        elif owner_type == "org" and response.get("organization", {}).get("projectV2"):
            formatted_response = {"user": response["organization"]}
        else:
            formatted_response = response

        return Project.from_response(formatted_response, self.client)

    def get_user(self, login: str) -> User:
        """Get information about a GitHub user.

        Args:
            login (str): GitHub username to fetch

        Returns:
            User: User object with profile information

        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_user")
        variables = {"login": login}
        response = self.client.execute(query, variables)
        return User.from_response(response)

    def get_issue(self, owner: str, repo: str, number: int) -> "Issue":
        """Get a specific issue by its repository and number.

        Args:
            owner (str): Repository owner (user or organization)
            repo (str): Repository name
            number (int): Issue number

        Returns:
            Issue: Issue object

        Raises:
            GraphQLError: On GraphQL operation failures
        """
        from .models.issue import Issue  # Lazy import

        query = load_query("get_issue")
        variables = {"owner": owner, "repo": repo, "number": number}
        response = self.client.execute(query, variables)
        return Issue.from_response(response, self.client, owner, repo)

    def get_organization(self, login: str) -> Organization:
        """Get information about a GitHub organization.

        Args:
            login (str): The organization's login/name

        Returns:
            Organization: Organization object with the fetched data

        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_organization")
        variables = {"login": login}
        response = self.client.execute(query, variables)
        return Organization.from_response(response)

    def get_repository(self, owner: str, name: str) -> Repository:
        """Get information about a GitHub repository.

        Args:
            owner (str): The repository owner's login (user or organization)
            name (str): The repository name

        Returns:
            Repository: Repository object with the fetched data

        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_repository")
        variables = {"owner": owner, "name": name}
        response = self.client.execute(query, variables)
        return Repository.from_response(response)

    def get_milestone(self, owner: str, repo: str, number: int) -> Milestone:
        """Get a specific milestone by its repository and number.

        Args:
            owner (str): Repository owner (user or organization)
            repo (str): Repository name
            number (int): Milestone number

        Returns:
            Milestone: Milestone object with the fetched data

        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_milestone")
        variables = {"owner": owner, "repo": repo, "number": number}
        response = self.client.execute(query, variables)
        return Milestone.from_response(response)

    def get_label(self, owner: str, repo: str, name: str) -> Label:
        """Get a specific label by its repository and name.

        Args:
            owner (str): Repository owner (user or organization)
            repo (str): Repository name
            name (str): Label name

        Returns:
            Label: Label object with the fetched data

        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_label")
        variables = {"owner": owner, "repo": repo, "name": name}
        response = self.client.execute(query, variables)
        return Label.from_response(response)

    def list_project_fields(self, project_id: str, first: int = 20) -> List["Field"]:
        """Get fields configured in a project.

        Args:
            project_id (str): The node ID of the project
            first (int, optional): Number of items per page. Defaults to 20.

        Returns:
            List[Field]: List of project field objects
        """
        from .models.field import Field  # Avoid circular import

        query = load_query("list_project_fields")
        variables = {"projectId": project_id, "first": first}
        result = _paginate_query(self.client, query, variables, "node.fields")
        field_nodes = result["node"]["fields"]["nodes"]
        
        # Convert raw field data to Field objects
        return [Field.from_response(field_data) for field_data in field_nodes]

    def search(
        self,
        query: str,
        search_type: Union[SearchType, str],
        first: int = 20,
        after: str = None,
    ) -> List[Any]:
        """Search GitHub using the GraphQL API.

        Args:
            query (str): Search query string (see GitHub's search syntax)
            search_type (Union[SearchType, str]): Type of items to search for
            first (int, optional): Number of results per page. Defaults to 20.
            after (str, optional): Cursor for pagination. Defaults to None.

        Returns:
            List[Any]: List of search result objects based on the search_type

        Raises:
            GraphQLError: On GraphQL operation failures
        """
        from .models.issue import Issue  # Lazy import
        from .models.pull_request import PullRequest  # Lazy import
        from .models.draft_issue import DraftIssue  # Lazy import
        from .models.repository import Repository  # Lazy import

        # Convert string type to enum if needed
        if isinstance(search_type, str):
            search_type = SearchType(search_type.upper())

        variables = {
            "query": query,
            "type": search_type,
            "first": first
        }
        if after:
            variables["after"] = after

        gql_query = load_query("search")
        response = _paginate_query(self.client, gql_query, variables, "search")
        nodes = response["search"]["nodes"]
        result = []

        # Transform nodes to appropriate type based on search_type
        for node in nodes:
            if search_type == SearchType.ISSUE:
                if node["__typename"] == "Issue":
                    owner, repo = node["repository"]["nameWithOwner"].split("/")
                    result.append(Issue.from_response(
                        {"node": node}, self.client, owner, repo))
                elif node["__typename"] == "PullRequest":
                    owner, repo = node["repository"]["nameWithOwner"].split("/")
                    result.append(PullRequest.from_response(
                        {"node": node}, self.client, owner, repo))
                elif node["__typename"] == "DraftIssue":
                    owner, repo = node["repository"]["nameWithOwner"].split("/")
                    result.append(DraftIssue.from_response(
                        {"node": node}, self.client, owner, repo))
            elif search_type == SearchType.REPOSITORY:
                if node["__typename"] == "Repository":
                    # Adapt node format to match what Repository.from_response expects
                    result.append(Repository.from_response({"repository": node}))
            else:
                # For unhandled types, append the raw node
                result.append(node)
        return result



"""ReferencedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class ReferencedEvent(Event):
    """Event representing an issue being referenced by another issue or pull request."""

    source_issue_number: Optional[int] = None
    source_issue_title: str = ""
    source_issue_url: Optional[str] = None
    source_repository_name: str = ""
    source_repository_owner: str = ""
    reference_type: str = ""

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "ReferencedEvent":
        """Create a ReferencedEvent instance from API response data."""
        event = super().from_response(data)

        source_issue_number = None
        source_issue_title = ""
        source_issue_url = None
        source_repository_name = ""
        source_repository_owner = ""
        reference_type = ""

        if data.get("subject") and data["subject"] is not None:
            source = data["subject"]
            reference_type = source.get("__typename", "")
            source_issue_number = source.get("number")
            source_issue_title = source.get("title", "")
            source_issue_url = source.get("url")

            if source.get("repository") and source["repository"] is not None:
                repository = source["repository"]
                source_repository_name = repository.get("name", "")

                if repository.get("owner") and repository["owner"] is not None:
                    source_repository_owner = repository["owner"].get("login", "")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            source_issue_number=source_issue_number,
            source_issue_title=source_issue_title,
            source_issue_url=source_issue_url,
            source_repository_name=source_repository_name,
            source_repository_owner=source_repository_owner,
            reference_type=reference_type,
        )



"""IssueTypeChangedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class IssueTypeChangedEvent(Event):
    """Event representing an issue type being changed."""

    prev_issue_type_id: Optional[str] = None
    prev_issue_type_name: Optional[str] = None
    prev_issue_type_description: Optional[str] = None
    prev_issue_type_color: Optional[str] = None

    issue_type_id: Optional[str] = None
    issue_type_name: Optional[str] = None
    issue_type_description: Optional[str] = None
    issue_type_color: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "IssueTypeChangedEvent":
        """Create an IssueTypeChangedEvent instance from API response data."""
        event = super().from_response(data)

        prev_issue_type_id = None
        prev_issue_type_name = None
        prev_issue_type_description = None
        prev_issue_type_color = None

        issue_type_id = None
        issue_type_name = None
        issue_type_description = None
        issue_type_color = None

        if data.get("prevIssueType") and data["prevIssueType"] is not None:
            prev_issue_type = data["prevIssueType"]
            prev_issue_type_id = prev_issue_type.get("id")
            prev_issue_type_name = prev_issue_type.get("name")
            prev_issue_type_description = prev_issue_type.get("description")
            prev_issue_type_color = prev_issue_type.get("color")

        if data.get("issueType") and data["issueType"] is not None:
            issue_type = data["issueType"]
            issue_type_id = issue_type.get("id")
            issue_type_name = issue_type.get("name")
            issue_type_description = issue_type.get("description")
            issue_type_color = issue_type.get("color")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            prev_issue_type_id=prev_issue_type_id,
            prev_issue_type_name=prev_issue_type_name,
            prev_issue_type_description=prev_issue_type_description,
            prev_issue_type_color=prev_issue_type_color,
            issue_type_id=issue_type_id,
            issue_type_name=issue_type_name,
            issue_type_description=issue_type_description,
            issue_type_color=issue_type_color,
        )



"""Custom exceptions for the GitHub GraphQL client."""


class GraphQLError(Exception):
    """Base exception for GitHub GraphQL API errors.
    
    Attributes:
        errors (list): List of GraphQL error dictionaries from the API response
        message (str): General error message
    """

    def __init__(self, message: str, errors=None):
        """Initialize the exception.

        Args:
            message (str): Error message
            errors (list, optional): List of GraphQL errors
        """
        super().__init__(message)
        self.errors = errors or []
    
    def __str__(self):
        """Return a string representation of the error.
        
        Returns:
            str: Formatted error message including API errors if available
        """
        if not self.errors:
            return super().__str__()
        
        error_messages = [f"- {error.get('message', 'Unknown error')}" for error in self.errors]
        return f"{super().__str__()}: \n" + "\n".join(error_messages)


class AuthenticationError(GraphQLError):
    """Raised when there are authentication issues."""

    pass


class RateLimitError(GraphQLError):
    """Raised when GitHub's rate limit is exceeded."""

    pass



"""MilestonedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class MilestonedEvent(Event):
    """Event representing an issue being added to a milestone."""

    milestone_title: Optional[str] = None
    subject_number: Optional[int] = None
    subject_title: Optional[str] = None
    subject_url: Optional[str] = None
    subject_repository_name: Optional[str] = None
    subject_repository_owner: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "MilestonedEvent":
        """Create a MilestonedEvent instance from API response data."""
        event = super().from_response(data)

        milestone_title = data.get("milestoneTitle")
        subject_number = None
        subject_title = None
        subject_url = None
        subject_repository_name = None
        subject_repository_owner = None

        if data.get("subject") and data["subject"] is not None:
            subject = data["subject"]
            subject_number = subject.get("number")
            subject_title = subject.get("title")
            subject_url = subject.get("url")

            if subject.get("repository"):
                subject_repository_name = subject["repository"].get("name")
                if subject["repository"].get("owner"):
                    subject_repository_owner = subject["repository"]["owner"].get(
                        "login"
                    )

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            milestone_title=milestone_title,
            subject_number=subject_number,
            subject_title=subject_title,
            subject_url=subject_url,
            subject_repository_name=subject_repository_name,
            subject_repository_owner=subject_repository_owner,
        )



"""Models representing GitHub Labels."""

from dataclasses import dataclass
from typing import Optional, Union, List
from datetime import datetime

from ..client import GraphQLClient
from ..queries import load_query
from .utils import parse_datetime

@dataclass
class Label:
    """Represents a GitHub label."""

    id: str = ""
    name: str = ""
    color: str = ""
    description: Optional[str] = None
    url: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    is_default: bool = False
    client: Optional[GraphQLClient] = None

    """Initialize a new Label.

    Args:
        id (str): The GitHub node ID
        name (str): The label name
        color (str): The hex color code without #
        description (Optional[str]): Label description text
        url (Optional[str]): URL to view the label
        created_at (Optional[datetime]): Creation timestamp
        updated_at (Optional[datetime]): Last update timestamp
        is_default (bool): Whether this is a default label
        client (Optional[GraphQLClient]): GraphQL client for API calls
    """

    @classmethod
    def from_response(cls, response: dict) -> Optional["Label"]:
        """Create a Label from a GraphQL response.

        Args:
            response (dict): The GraphQL response data

        Returns:
            Optional[Label]: A Label object, or None if invalid data
        """
        repo_data = response.get("repository", {})
        label_data = repo_data.get("label")

        if not label_data:
            # Check if this is a node response (for label by ID)
            node_data = response.get("node")
            if node_data:
                label_data = node_data
            else:
                return None

        return cls(
            id=label_data["id"],
            name=label_data["name"],
            color=label_data["color"],
            description=label_data.get("description"),
            url=label_data.get("url"),
            created_at=parse_datetime(label_data.get("createdAt")),
            updated_at=parse_datetime(label_data.get("updatedAt")),
            is_default=label_data.get("isDefault", False),
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        repository: str,
        name: Optional[str] = None,
        id: Optional[str] = None,
        owner: Optional[str] = None
    ) -> "Label":
        """Get a label's details and populate this instance.
        
        If name is provided, get all labels on the repo and search for one that matches.
        If id is provided, query directly for the Label with that ID.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            repository (str): Repository name
            name (Optional[str]): Label name
            id (Optional[str]): Label node ID
            owner (Optional[str]): Repository owner (user or organization)
            
        Returns:
            Label: Self with populated properties
            
        Raises:
            ValueError: If neither name nor id is provided, or if owner is missing when using name
            ValueError: If label with provided name is not found
            GraphQLError: On GraphQL operation failures
        """
        if not name and not id:
            raise ValueError("Either name or id must be provided")
            
        self.client = client
        
        if id:
            # Get label by ID
            query = load_query("get_label_by_id")
            variables = {"id": id}
            
            response = client.execute(query, variables)
            label = Label.from_response(response)
            
            if not label:
                raise ValueError(f"Label with ID {id} not found")
        else:
            # Get label by name
            if not owner:
                raise ValueError("Owner is required when getting label by name")
                
            query = load_query("get_label")
            variables = {
                "owner": owner,
                "repo": repository,
                "name": name
            }
            
            response = client.execute(query, variables)
            label = Label.from_response(response)
            
            if not label:
                raise ValueError(f"Label '{name}' not found in {owner}/{repository}")
        
        # Update this instance's attributes with the fetched data
        self.id = label.id
        self.name = label.name
        self.color = label.color
        self.description = label.description
        self.url = label.url
        self.created_at = label.created_at
        self.updated_at = label.updated_at
        self.is_default = label.is_default
        
        return self


"""MentionedEvent model class."""

from typing import Dict, Any
from dataclasses import dataclass
from .event import Event


@dataclass
class MentionedEvent(Event):
    """Event representing a user being mentioned in an issue."""

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "MentionedEvent":
        """Create a MentionedEvent instance from API response data."""
        event = super().from_response(data)

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
        )



"""AddedToProjectEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class AddedToProjectEvent(Event):
    """Event representing an issue being added to a project."""

    project_name: str = ""
    project_url: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "AddedToProjectEvent":
        """Create an AddedToProjectEvent instance from API response data."""
        event = super().from_response(data)

        project_name = ""
        project_url = None
        if data.get("project") and data["project"] is not None:
            project_url = data["project"].get("url")
            project_name = data["project"].get("name", "")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            project_name=project_name,
            project_url=project_url,
        )



"""IssueCommentEvent model class."""

from typing import Dict, Any
from dataclasses import dataclass, field
from datetime import datetime
from .event import Event


@dataclass
class IssueCommentEvent(Event):
    """Represents a comment on a GitHub issue."""

    body: str = ""
    url: str = ""
    updated_at: datetime = field(default_factory=lambda: datetime.now())

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "IssueCommentEvent":
        """Create an IssueCommentEvent instance from API response data."""
        event = super().from_response(data)

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            body=data.get("body", ""),
            url=data.get("url", ""),
            updated_at=datetime.fromisoformat(data["updatedAt"].replace("Z", "+00:00")),
        )



"""Project PullRequest model class."""

from typing import Dict, Any, List, Optional
from dataclasses import dataclass
from datetime import datetime
from ..client import GraphQLClient
from ..queries import load_query
from .comment import Comment
from .events.event import Event


@dataclass
class PullRequest:
    """Represents a pull request in a GitHub Project (V2)."""

    id: str
    number: int
    title: str
    url: str
    state: str
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    field_values: List[Dict[str, Any]] = None
    author_login: str = ""
    body: Optional[str] = None
    client: Optional[GraphQLClient] = None
    owner: Optional[str] = None
    repo: Optional[str] = None

    def __post_init__(self):
        """Initialize default values for datetime fields."""
        if self.created_at is None:
            self.created_at = datetime.now()
        if self.updated_at is None:
            self.updated_at = datetime.now()
        if self.field_values is None:
            self.field_values = []

    @classmethod
    def from_response(
        cls,
        data: Dict[str, Any],
        client: Optional[GraphQLClient] = None,
        owner: Optional[str] = None,
        repo: Optional[str] = None,
    ) -> "PullRequest":
        """Create a PullRequest instance from API response data.

        Args:
            data (Dict[str, Any]): Raw pull request data from API response
            client (Optional[GraphQLClient]): GraphQL client for API calls
            owner (Optional[str]): Owner of the repository
            repo (Optional[str]): Repository name

        Returns:
            PullRequest: New PullRequest instance
        """
        if "content" in data:  # Project item response
            content = data["content"]
            return cls(
                id=content["id"],
                number=content["number"], 
                title=content["title"],
                url=content["url"],
                state=content["state"],
                created_at=datetime.fromisoformat(
                    content["createdAt"].replace("Z", "+00:00")
                ),
                updated_at=datetime.fromisoformat(
                    content["updatedAt"].replace("Z", "+00:00")
                ),
                author_login=content.get("author", {}).get("login", ""),
                body=None,  # Not included in project responses
                field_values=data.get("fieldValues", {"nodes": []})["nodes"],
                client=client,
                owner=owner,
                repo=repo,
            )
        else:  # PR-specific response or search response
            # For search results it's in ["node"], for direct access in data, for specific queries under repository
            if "node" in data:
                pr_data = data["node"]
            elif "repository" in data:
                pr_data = data["repository"]["pullRequest"]
            else:
                pr_data = data
            
            return cls(
                id=pr_data["id"],
                number=pr_data["number"],
                title=pr_data["title"],
                url=pr_data["url"],
                state=pr_data["state"],
                created_at=datetime.fromisoformat(
                    pr_data["createdAt"].replace("Z", "+00:00")
                ),
                updated_at=datetime.fromisoformat(
                    pr_data["updatedAt"].replace("Z", "+00:00")
                ),
                author_login=pr_data["author"]["login"],
                body=pr_data.get("body"),
                field_values=[],  # Not included in PR responses
                client=client,
                owner=owner,
                repo=repo,
            )

    def get_comments(self, first: int = 20) -> Dict[str, List[Comment]]:
        """Get comments on this pull request.

        Args:
            first (int, optional): Number of comments to fetch per page. Defaults to 20.

        Returns:
            dict: Pull request comments as Comment objects, in "nodes" key

        Raises:
            ValueError: If client, owner or repo are not set
            GraphQLError: On GraphQL operation failures
        """
        if not all([self.client, self.owner, self.repo]):
            raise ValueError("client, owner and repo must be set to fetch comments")

        query = load_query("get_issue_comments")  # PRs use same query structure as issues
        variables = {
            "owner": self.owner,
            "repo": self.repo,
            "number": self.number,
            "first": first,
        }

        response = self.client.execute(query, variables)
        comments_data = response["repository"]["issue"]["comments"]
        comments_data["nodes"] = [
            Comment.from_response(node) for node in comments_data["nodes"]
        ]
        return comments_data
        
    def get(
        self, 
        *,
        client: GraphQLClient, 
        repository: str,
        number: int,
        org: Optional[str] = None,
        username: Optional[str] = None
    ) -> "PullRequest":
        """Get a pull request's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            repository (str): Repository name
            number (int): Pull request number
            org (Optional[str]): Organization name (if PR is in an org's repo)
            username (Optional[str]): Username (if PR is in a user's repo)
            
        Returns:
            PullRequest: Self with populated properties
            
        Raises:
            ValueError: If neither org nor username is provided
            GraphQLError: On GraphQL operation failures
        """
        if not org and not username:
            raise ValueError("Either org or username must be provided")
            
        owner = org if org else username
        query = load_query("get_pull_request")
        variables = {
            "owner": owner,
            "repo": repository,
            "number": number
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create a pull request instance with all properties
        pr = PullRequest.from_response(response, client, owner, repository)
        
        # Update this instance's attributes with the fetched data
        self.id = pr.id
        self.number = pr.number
        self.title = pr.title
        self.url = pr.url
        self.state = pr.state
        self.created_at = pr.created_at
        self.updated_at = pr.updated_at
        self.author_login = pr.author_login
        self.body = pr.body
        self.field_values = pr.field_values
        self.client = client
        self.owner = owner
        self.repo = repository
        
        return self
        
    def get_timeline(self, first: int = 20) -> List[Event]:
        """Get timeline events on this pull request.

        Args:
            first (int, optional): Number of events to fetch per page. Defaults to 20.

        Returns:
            List[Event]: Pull request timeline events as Event objects

        Raises:
            ValueError: If client, owner or repo are not set
            GraphQLError: On GraphQL operation failures
        """
        if not all([self.client, self.owner, self.repo]):
            raise ValueError("client, owner and repo must be set to fetch timeline events")

        query = load_query("get_pull_request_timeline")
        variables = {
            "owner": self.owner,
            "repo": self.repo,
            "number": self.number,
            "first": first,
        }

        response = self.client.execute(query, variables)
        timeline_data = response["repository"]["pullRequest"]["timelineItems"]
        nodes = [
            Event.create_from_response(node) for node in timeline_data["nodes"] if Event.create_from_response(node)
        ]
        return nodes


"""Organization model class for GitHub API responses."""

from dataclasses import dataclass
from typing import Optional, Dict, List, Any

from ..client import GraphQLClient
from ..queries import load_query
from .repository import Repository
from .user import User


@dataclass
class Organization:
    """Represents a GitHub organization."""

    id: str
    login: str
    name: Optional[str] = None
    description: Optional[str] = None
    url: Optional[str] = None
    location: Optional[str] = None
    website_url: Optional[str] = None
    client: Optional[GraphQLClient] = None

    @classmethod
    def from_response(cls, response: dict, client: Optional[GraphQLClient] = None) -> "Organization":
        """Create Organization instance from GraphQL response.

        Args:
            response (dict): GraphQL response data containing organization info
            client (Optional[GraphQLClient]): GraphQL client for API calls

        Returns:
            Organization: New Organization instance
        """
        # For node query the data is in ["node"], for direct query in ["organization"]
        if "node" in response:
            org_data = response["node"]
        else:
            org_data = response.get("organization", {})
            
        return cls(
            id=org_data["id"],
            login=org_data["login"],
            name=org_data.get("name"),
            description=org_data.get("description"),
            url=org_data.get("url"),
            location=org_data.get("location"),
            website_url=org_data.get("websiteUrl"),
            client=client,
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        id: Optional[str] = None,
        name: Optional[str] = None
    ) -> "Organization":
        """Get an organization's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            id (Optional[str]): Organization ID (for lookup by ID)
            name (Optional[str]): Organization name/login (for lookup by name)
            
        Returns:
            Organization: Self with populated properties
            
        Raises:
            ValueError: If neither id nor name is provided
            GraphQLError: On GraphQL operation failures
        """
        if not id and not name:
            raise ValueError("Either id or name must be provided")
            
        if id:
            query = load_query("get_organization_by_id")
            variables = {"id": id}
        else:
            query = load_query("get_organization")
            variables = {"login": name}
        
        response = client.execute(query, variables)
        
        # Use from_response to create an organization instance with all properties
        org = Organization.from_response(response, client)
        
        # Update this instance's attributes with the fetched data
        self.id = org.id
        self.login = org.login
        self.name = org.name
        self.description = org.description
        self.url = org.url
        self.location = org.location
        self.website_url = org.website_url
        self.client = client
        
        return self
        
    def get_repos(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get repositories belonging to this organization.
        
        Args:
            first (int, optional): Number of repositories to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of Repository objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set or login is not set
            GraphQLError: On GraphQL operation failures
        """
        if not self.client:
            raise ValueError("Client must be set to fetch repositories")
        
        if not self.login:
            raise ValueError("Organization login must be set to fetch repositories")
            
        query = load_query("get_organization_repos")
        variables = {
            "login": self.login,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        repos_data = response["organization"]["repositories"]
        
        # Convert each node to a Repository object
        repos_data["nodes"] = [
            Repository.from_response({"repository": repo}) for repo in repos_data["nodes"]
        ]
        
        return repos_data
        
    def get_users(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get users (members) belonging to this organization.
        
        Args:
            first (int, optional): Number of users to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of User objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set or login is not set
            GraphQLError: On GraphQL operation failures
        """
        if not self.client:
            raise ValueError("Client must be set to fetch users")
        
        if not self.login:
            raise ValueError("Organization login must be set to fetch users")
            
        query = load_query("get_organization_users")
        variables = {
            "login": self.login,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        users_data = response["organization"]["membersWithRole"]
        
        # Convert each node to a User object
        users_data["nodes"] = [
            User.from_response({"user": user}, self.client) for user in users_data["nodes"]
        ]
        
        return users_data



"""Common utilities for model classes."""

from datetime import datetime, timezone


def parse_datetime(date_str: str | None) -> datetime | None:
    """Parse ISO format datetime string to datetime object with UTC timezone.

    Args:
        date_str (str | None): ISO format datetime string or None

    Returns:
        datetime | None: Parsed datetime with UTC timezone or None if input is None
    """
    if not date_str:
        return None
    return datetime.strptime(date_str, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)


"""MergedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class MergedEvent(Event):
    """Event representing a pull request being merged."""

    merge_commit_sha: Optional[str] = None
    merge_ref_name: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "MergedEvent":
        """Create a MergedEvent instance from API response data."""
        event = super().from_response(data)

        merge_commit_sha = (
            data.get("commit", {}).get("oid") if data.get("commit") else None
        )
        merge_ref_name = data.get("mergeRefName")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            merge_commit_sha=merge_commit_sha,
            merge_ref_name=merge_ref_name,
        )



"""Base Event model class for timeline events."""

from typing import Dict, Any, Optional, Type
from dataclasses import dataclass
from datetime import datetime


@dataclass
class Event:
    """Base class for all timeline events."""

    id: str
    created_at: datetime
    actor_login: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "Event":
        """Create an Event instance from API response data.

        Args:
            data (Dict[str, Any]): Raw event data from API response

        Returns:
            Event: New Event instance
        """
        actor_login = None
        if data.get("actor") and data["actor"] is not None:
            actor_login = data["actor"].get("login")

        return cls(
            id=data["id"],
            created_at=datetime.fromisoformat(data["createdAt"].replace("Z", "+00:00")),
            actor_login=actor_login,
        )

    @staticmethod
    def create_from_response(data: Dict[str, Any]) -> Optional["Event"]:
        """Factory method to create the appropriate event type from response data.

        Args:
            data (Dict[str, Any]): Raw event data from API response

        Returns:
            Optional[Event]: Event instance of the appropriate subclass, or None if type not supported
        """
        from . import (
            AddedToProjectEvent,
            AssignedEvent,
            ClosedEvent,
            CommentDeletedEvent,
            CrossReferencedEvent,
            DemilestonedEvent,
            IssueCommentEvent,
            IssueTypeAddedEvent,
            IssueTypeChangedEvent,
            IssueTypeRemovedEvent,
            LabeledEvent,
            LockedEvent,
            MarkedAsDuplicateEvent,
            MentionedEvent,
            MergedEvent,
            MilestonedEvent,
            MovedColumnsInProjectEvent,
            ParentIssueAddedEvent,
            ParentIssueRemovedEvent,
            PinnedEvent,
            ReferencedEvent,
            RenamedTitleEvent,
            ReopenedEvent,
            SubIssueAddedEvent,
            SubIssueRemovedEvent,
            SubscribedEvent,
            TransferredEvent,
            UnassignedEvent,
            UnlabeledEvent,
            UnsubscribedEvent,
        )

        event_type = data.get("__typename")
        event_classes = {
            "AddedToProjectEvent": AddedToProjectEvent,
            "AssignedEvent": AssignedEvent,
            "ClosedEvent": ClosedEvent,
            "CommentDeletedEvent": CommentDeletedEvent,
            "CrossReferencedEvent": CrossReferencedEvent,
            "DemilestonedEvent": DemilestonedEvent,
            "IssueComment": IssueCommentEvent,
            "IssueTypeAddedEvent": IssueTypeAddedEvent,
            "IssueTypeChangedEvent": IssueTypeChangedEvent,
            "IssueTypeRemovedEvent": IssueTypeRemovedEvent,
            "LabeledEvent": LabeledEvent,
            "LockedEvent": LockedEvent,
            "MarkedAsDuplicateEvent": MarkedAsDuplicateEvent,
            "MentionedEvent": MentionedEvent,
            "MergedEvent": MergedEvent,
            "MilestonedEvent": MilestonedEvent,
            "MovedColumnsInProjectEvent": MovedColumnsInProjectEvent,
            "ParentIssueAddedEvent": ParentIssueAddedEvent,
            "ParentIssueRemovedEvent": ParentIssueRemovedEvent,
            "PinnedEvent": PinnedEvent,
            "ReferencedEvent": ReferencedEvent,
            "RenamedTitleEvent": RenamedTitleEvent,
            "ReopenedEvent": ReopenedEvent,
            "SubIssueAddedEvent": SubIssueAddedEvent,
            "SubIssueRemovedEvent": SubIssueRemovedEvent,
            "SubscribedEvent": SubscribedEvent,
            "TransferredEvent": TransferredEvent,
            "UnassignedEvent": UnassignedEvent,
            "UnlabeledEvent": UnlabeledEvent,
            "UnsubscribedEvent": UnsubscribedEvent,
        }

        if event_type in event_classes:
            return event_classes[event_type].from_response(data)
        return None



"""General utility functions for the gh-project-v2 library."""

import os
from datetime import datetime
from pathlib import Path


def log_message(message: str) -> None:
    """Log a message to a file if LOG_FILE environment variable is set.

    This function writes log messages to a file specified by the LOG_FILE
    environment variable. If the variable is not set, no logging occurs.

    The log format is: [%Y-%m-%d %H:%M:%S] <log message>
    Each log entry is separated by newlines.

    Args:
        message (str): The message to log

    Example:
        >>> import os
        >>> os.environ['LOG_FILE'] = '/tmp/test.log'
        >>> log_message("This is a test message")
        # Writes: [2024-01-01 12:00:00] This is a test message
    """
    log_file_path = os.environ.get('LOG_FILE')

    if not log_file_path:
        # No logging if LOG_FILE is not set
        return

    # Format the timestamp
    timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    log_entry = f"[{timestamp}] {message}\n"

    try:
        # Create parent directories if they don't exist
        log_path = Path(log_file_path)
        log_path.parent.mkdir(parents=True, exist_ok=True)

        # Append the log entry to the file
        with open(log_file_path, 'a', encoding='utf-8') as f:
            f.write(log_entry)
    except (OSError, IOError):
        # Silently ignore logging errors to avoid disrupting the main application
        # In a production environment, you might want to use a fallback logging mechanism
        pass



"""MovedColumnsInProjectEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class MovedColumnsInProjectEvent(Event):
    """Event representing an issue being moved between columns in a project."""

    project_name: str = ""
    project_url: Optional[str] = None
    previous_column_name: Optional[str] = None
    current_column_name: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "MovedColumnsInProjectEvent":
        """Create a MovedColumnsInProjectEvent instance from API response data."""
        event = super().from_response(data)

        project_name = ""
        project_url = None
        previous_column_name = None
        current_column_name = None

        if data.get("project") and data["project"] is not None:
            project_url = data["project"].get("url")
            project_name = data["project"].get("name", "")

        if data.get("previousProjectColumnName"):
            previous_column_name = data["previousProjectColumnName"]

        if data.get("projectColumnName"):
            current_column_name = data["projectColumnName"]

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            project_name=project_name,
            project_url=project_url,
            previous_column_name=previous_column_name,
            current_column_name=current_column_name,
        )



"""Issue Timeline Event model classes - compatibility import."""

# Import all event classes from the new location for backward compatibility
from .events import (
    Event, AddedToProjectEvent, AssignedEvent, ClosedEvent, CommentDeletedEvent,
    CrossReferencedEvent, IssueCommentEvent, IssueTypeAddedEvent, IssueTypeChangedEvent,
    LabeledEvent, MentionedEvent, RenamedTitleEvent, SubIssueAddedEvent, 
    SubIssueRemovedEvent, TransferredEvent
)

__all__ = [
    "Event", "AddedToProjectEvent", "AssignedEvent", "ClosedEvent", "CommentDeletedEvent",
    "CrossReferencedEvent", "IssueCommentEvent", "IssueTypeAddedEvent", "IssueTypeChangedEvent",
    "LabeledEvent", "MentionedEvent", "RenamedTitleEvent", "SubIssueAddedEvent", 
    "SubIssueRemovedEvent", "TransferredEvent"
]


"""Project View model class."""

from typing import Dict, Any, List, Optional
from dataclasses import dataclass

from ..client import GraphQLClient
from ..queries import load_query


@dataclass
class View:
    """Represents a view in a GitHub Project (V2)."""

    id: str
    name: str
    number: int = 0
    layout: str = ""
    fields: List[Dict[str, Any]] = None
    client: Optional[GraphQLClient] = None
    project_id: Optional[str] = None
    
    def __post_init__(self):
        """Initialize default values for list fields."""
        if self.fields is None:
            self.fields = []

    @classmethod
    def from_response(cls, data: Dict[str, Any], client: Optional[GraphQLClient] = None, project_id: Optional[str] = None) -> "View":
        """Create a View instance from API response data.

        Args:
            data (Dict[str, Any]): Raw view data from API response
            client (Optional[GraphQLClient]): GraphQL client for API calls
            project_id (Optional[str]): ID of the project this view belongs to

        Returns:
            View: New View instance
        """
        # For node query, data is in ["node"]
        if "node" in data:
            view_data = data["node"]
        else:
            view_data = data
            
        return cls(
            id=view_data["id"],
            name=view_data["name"],
            number=view_data["number"],
            layout=view_data["layout"],
            fields=view_data["fields"]["nodes"],
            client=client,
            project_id=project_id,
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        id: Optional[str] = None,
        name: Optional[str] = None,
        project_id: Optional[str] = None
    ) -> "View":
        """Get a view's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            id (Optional[str]): View ID (for lookup by ID)
            name (Optional[str]): View name (for lookup by name, requires project_id)
            project_id (Optional[str]): Project ID (required when looking up by name)
            
        Returns:
            View: Self with populated properties
            
        Raises:
            ValueError: If neither id nor (name and project_id) are provided
            GraphQLError: On GraphQL operation failures
        """
        if not id and not (name and project_id):
            raise ValueError("Either id or (name and project_id) must be provided")
            
        if id:
            query = load_query("get_view")
            variables = {
                "viewId": id,
                "fieldsFirst": 100
            }
            
            response = client.execute(query, variables)
            
            # Use from_response to create a view instance with all properties
            view = View.from_response(response, client)
            
            # Update this instance's attributes with the fetched data
            self.id = view.id
            self.name = view.name
            self.number = view.number
            self.layout = view.layout
            self.fields = view.fields
            self.client = client
            self.project_id = None  # Not returned in the query
            
        else:
            # To find by name, we need to get all views from the project and filter
            query = load_query("get_project_views")
            variables = {
                "projectId": project_id,
                "viewsFirst": 20  # Adjust if needed
            }
            
            response = client.execute(query, variables)
            
            # Find the view with the matching name
            project_views = response["node"]["views"]["nodes"]
            matching_views = [v for v in project_views if v["name"] == name]
            
            if not matching_views:
                raise ValueError(f"No view found with name '{name}' in project")
                
            view_data = matching_views[0]
            
            # We need to fetch the complete view data with the ID
            view_id = view_data["id"]
            return self.get(client=client, id=view_id)
            
        return self



"""Workflow model class for GitHub API responses."""

from dataclasses import dataclass
from typing import Optional, Dict, Any, List

from ..client import GraphQLClient
from ..queries import load_query


@dataclass(eq=True, frozen=True)
class Workflow:
    """Represents a GitHub workflow."""

    id: str
    name: str
    path: str
    state: str
    url: Optional[str] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    file_contents: Optional[str] = None
    ref: Optional[str] = None
    run_count: int = 0
    client: Optional[GraphQLClient] = None

    def __hash__(self):
        """Make Workflow hashable by its id."""
        return hash(self.id)

    @classmethod
    def from_response(cls, data: Dict[str, Any], client: Optional[GraphQLClient] = None) -> "Workflow":
        """Create Workflow instance from GraphQL response.

        Args:
            data (Dict[str, Any]): GraphQL response data containing workflow info
            client (Optional[GraphQLClient], optional): GraphQL client for API calls. Defaults to None.

        Returns:
            Workflow: New Workflow instance
        """
        workflow_data = data.get("node", data.get("workflow", {}))
        
        return cls(
            id=workflow_data.get("id", ""),
            name=workflow_data.get("name", ""),
            path=workflow_data.get("path", ""),
            state=workflow_data.get("state", ""),
            url=workflow_data.get("url"),
            created_at=workflow_data.get("createdAt"),
            updated_at=workflow_data.get("updatedAt"),
            file_contents=workflow_data.get("fileContents"),
            ref=workflow_data.get("ref"),
            run_count=workflow_data.get("runCount", 0),
            client=client,
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        id: str,
        owner: Optional[str] = None,
        repo: Optional[str] = None
    ) -> "Workflow":
        """Get a workflow's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            id (str): Workflow node ID
            owner (Optional[str]): Repository owner (organization or username)
            repo (Optional[str]): Repository name
            
        Returns:
            Workflow: Self with populated properties
            
        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_workflow")
        variables = {
            "id": id,
            "owner": owner or "",
            "repo": repo or ""
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create a workflow instance with all properties
        workflow = Workflow.from_response(response, client)
        
        # Update this instance's attributes with the fetched data
        object.__setattr__(self, "id", workflow.id)
        object.__setattr__(self, "name", workflow.name)
        object.__setattr__(self, "path", workflow.path)
        object.__setattr__(self, "state", workflow.state)
        object.__setattr__(self, "url", workflow.url)
        object.__setattr__(self, "created_at", workflow.created_at)
        object.__setattr__(self, "updated_at", workflow.updated_at)
        object.__setattr__(self, "file_contents", workflow.file_contents)
        object.__setattr__(self, "ref", workflow.ref)
        object.__setattr__(self, "run_count", workflow.run_count)
        object.__setattr__(self, "client", client)
        
        return self


"""UnlabeledEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class UnlabeledEvent(Event):
    """Event representing a label being removed from an issue."""

    label_name: str = ""
    label_color: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "UnlabeledEvent":
        """Create an UnlabeledEvent instance from API response data."""
        event = super().from_response(data)

        label_name = ""
        label_color = None
        if data.get("label") and data["label"] is not None:
            label_name = data["label"].get("name", "")
            label_color = data["label"].get("color")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            label_name=label_name,
            label_color=label_color,
        )



"""Issue Timeline Event model classes."""

from .event import Event
from .added_to_project_event import AddedToProjectEvent
from .assigned_event import AssignedEvent
from .closed_event import ClosedEvent
from .comment_deleted_event import CommentDeletedEvent
from .cross_referenced_event import CrossReferencedEvent
from .demilestoned_event import DemilestonedEvent
from .issue_comment_event import IssueCommentEvent
from .issue_type_added_event import IssueTypeAddedEvent
from .issue_type_changed_event import IssueTypeChangedEvent
from .issue_type_removed_event import IssueTypeRemovedEvent
from .labeled_event import LabeledEvent
from .locked_event import LockedEvent
from .marked_as_duplicate_event import MarkedAsDuplicateEvent
from .mentioned_event import MentionedEvent
from .merged_event import MergedEvent
from .milestoned_event import MilestonedEvent
from .moved_columns_in_project_event import MovedColumnsInProjectEvent
from .parent_issue_added_event import ParentIssueAddedEvent
from .parent_issue_removed_event import ParentIssueRemovedEvent
from .pinned_event import PinnedEvent
from .referenced_event import ReferencedEvent
from .renamed_title_event import RenamedTitleEvent
from .reopened_event import ReopenedEvent
from .sub_issue_added_event import SubIssueAddedEvent
from .sub_issue_removed_event import SubIssueRemovedEvent
from .subscribed_event import SubscribedEvent
from .transferred_event import TransferredEvent
from .unassigned_event import UnassignedEvent
from .unlabeled_event import UnlabeledEvent
from .unsubscribed_event import UnsubscribedEvent

__all__ = [
    "Event",
    "AddedToProjectEvent",
    "AssignedEvent",
    "ClosedEvent",
    "CommentDeletedEvent",
    "CrossReferencedEvent",
    "DemilestonedEvent",
    "IssueCommentEvent",
    "IssueTypeAddedEvent",
    "IssueTypeChangedEvent",
    "IssueTypeRemovedEvent",
    "LabeledEvent",
    "LockedEvent",
    "MarkedAsDuplicateEvent",
    "MentionedEvent",
    "MergedEvent",
    "MilestonedEvent",
    "MovedColumnsInProjectEvent",
    "ParentIssueAddedEvent",
    "ParentIssueRemovedEvent",
    "PinnedEvent",
    "ReferencedEvent",
    "RenamedTitleEvent",
    "ReopenedEvent",
    "SubIssueAddedEvent",
    "SubIssueRemovedEvent",
    "SubscribedEvent",
    "TransferredEvent",
    "UnassignedEvent",
    "UnlabeledEvent",
    "UnsubscribedEvent",
]



"""Issue Comment model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from datetime import datetime

from ..client import GraphQLClient
from ..queries import load_query


@dataclass
class Comment:
    """Represents a comment on a GitHub issue."""

    id: str
    body: str
    author_login: str
    url: str
    created_at: datetime
    updated_at: datetime
    client: Optional[GraphQLClient] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any], client: Optional[GraphQLClient] = None) -> "Comment":
        """Create a Comment instance from API response data.

        Args:
            data (Dict[str, Any]): Raw comment data from API response
            client (Optional[GraphQLClient]): GraphQL client for API calls

        Returns:
            Comment: New Comment instance
        """
        return cls(
            id=data["id"],
            body=data["body"],
            author_login=data["author"]["login"],
            url=data["url"],
            created_at=datetime.fromisoformat(data["createdAt"].replace("Z", "+00:00")),
            updated_at=datetime.fromisoformat(data["updatedAt"].replace("Z", "+00:00")),
            client=client,
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        id: str
    ) -> "Comment":
        """Get a comment's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            id (str): Comment node ID
            
        Returns:
            Comment: Self with populated properties
            
        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_comment")
        variables = {
            "id": id
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create a comment instance with all properties
        comment = Comment.from_response(response["node"], client)
        
        # Update this instance's attributes with the fetched data
        self.id = comment.id
        self.body = comment.body
        self.author_login = comment.author_login
        self.url = comment.url
        self.created_at = comment.created_at
        self.updated_at = comment.updated_at
        self.client = client
        
        return self



"""ParentIssueAddedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class ParentIssueAddedEvent(Event):
    """Event representing a parent issue being added to an issue."""

    parent_issue_number: Optional[int] = None
    parent_issue_title: str = ""
    parent_issue_url: Optional[str] = None
    parent_repository_name: str = ""
    parent_repository_owner: str = ""

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "ParentIssueAddedEvent":
        """Create a ParentIssueAddedEvent instance from API response data."""
        event = super().from_response(data)

        parent_issue_number = None
        parent_issue_title = ""
        parent_issue_url = None
        parent_repository_name = ""
        parent_repository_owner = ""

        if data.get("parent") and data["parent"] is not None:
            parent_issue = data["parent"]
            parent_issue_number = parent_issue.get("number")
            parent_issue_title = parent_issue.get("title", "")
            parent_issue_url = parent_issue.get("url")

            if (
                parent_issue.get("repository")
                and parent_issue["repository"] is not None
            ):
                repository = parent_issue["repository"]
                parent_repository_name = repository.get("name", "")

                if repository.get("owner") and repository["owner"] is not None:
                    parent_repository_owner = repository["owner"].get("login", "")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            parent_issue_number=parent_issue_number,
            parent_issue_title=parent_issue_title,
            parent_issue_url=parent_issue_url,
            parent_repository_name=parent_repository_name,
            parent_repository_owner=parent_repository_owner,
        )



"""GitHub Projects V2 Models."""

from .view import View
from .issue import Issue
from .draft_issue import DraftIssue
from .project import Project
from .user import User
from .comment import Comment
from .organization import Organization
from .repository import Repository
from .label import Label
from .milestone import Milestone
from .pull_request import PullRequest
from .field import Field
from .option import Option
from .team import Team
from .workflow import Workflow
from .events import (
    Event, AddedToProjectEvent, AssignedEvent, ClosedEvent, CommentDeletedEvent,
    CrossReferencedEvent, IssueCommentEvent, IssueTypeAddedEvent, IssueTypeChangedEvent,
    LabeledEvent, MentionedEvent
)

__all__ = ["View", "Issue", "DraftIssue", "Project", "User", "Comment", "Organization", 
           "Repository", "Label", "Milestone", "PullRequest", "Field", "Option",
           "Team", "Workflow", "Event", "AddedToProjectEvent", "AssignedEvent", 
           "ClosedEvent", "CommentDeletedEvent", "CrossReferencedEvent", "IssueCommentEvent", 
           "IssueTypeAddedEvent", "IssueTypeChangedEvent", "LabeledEvent", "MentionedEvent"]



"""Repository model class for GitHub API responses."""

from dataclasses import dataclass
from typing import Optional, Dict, List, Any
import copy

from ..client import GraphQLClient
from ..queries import load_query
from .label import Label
from .pull_request import PullRequest


@dataclass(eq=True, frozen=True)
class Repository:
    """Represents a GitHub repository."""

    id: str = ""
    name: str = ""
    name_with_owner: str = ""
    url: str = ""
    owner_login: str = ""
    owner_url: str = ""
    description: Optional[str] = None
    is_private: bool = False
    homepage_url: Optional[str] = None
    client: Optional[GraphQLClient] = None

    def __hash__(self):
        """Make Repository hashable by its id."""
        return hash(self.id)

    @classmethod
    def from_response(cls, response: dict) -> "Repository":
        """Create Repository instance from GraphQL response.

        Args:
            response (dict): GraphQL response data containing repository info

        Returns:
            Repository: New Repository instance
        """
        repo_data = response.get("repository", {})
        owner = repo_data.get("owner", {})

        return cls(
            id=repo_data["id"],
            name=repo_data["name"],
            name_with_owner=repo_data["nameWithOwner"],
            url=repo_data["url"],
            owner_login=owner["login"],
            owner_url=owner["url"],
            description=repo_data.get("description"),
            is_private=repo_data.get("isPrivate", False),
            homepage_url=repo_data.get("homepageUrl"),
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        name: str,
        org: Optional[str] = None,
        username: Optional[str] = None
    ) -> "Repository":
        """Get a repository's details.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            name (str): Repository name
            org (Optional[str]): Organization name (if repository belongs to an org)
            username (Optional[str]): Username (if repository belongs to a user)
            
        Returns:
            Repository: New Repository instance with populated properties
            
        Raises:
            ValueError: If neither org nor username is provided
            GraphQLError: On GraphQL operation failures
        """
        if not org and not username:
            raise ValueError("Either org or username must be provided")
            
        owner = org if org else username
        query = load_query("get_repository")
        variables = {
            "owner": owner,
            "name": name
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create a repository instance with all properties
        repo = Repository.from_response(response)
        
        # Create a new Repository instance with the client set
        return Repository(
            id=repo.id,
            name=repo.name,
            name_with_owner=repo.name_with_owner,
            url=repo.url,
            owner_login=repo.owner_login,
            owner_url=repo.owner_url,
            description=repo.description,
            is_private=repo.is_private,
            homepage_url=repo.homepage_url,
            client=client
        )
        
    def get_labels(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get labels for this repository.
        
        Args:
            first (int, optional): Number of labels to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of Label objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set or owner_login/name are not set
            GraphQLError: On GraphQL operation failures
        """
        if not self.client:
            raise ValueError("Client must be set to fetch labels")
        
        if not self.owner_login or not self.name:
            raise ValueError("Repository owner_login and name must be set to fetch labels")
            
        query = load_query("get_repository_labels")
        variables = {
            "owner": self.owner_login,
            "repo": self.name,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        labels_data = response["repository"]["labels"]
        
        # Convert each node to a Label object
        labels_data["nodes"] = [
            Label(
                id=label["id"],
                name=label["name"],
                color=label["color"],
                description=label.get("description"),
                url=label.get("url"),
                created_at=Label.from_response({"repository": {"label": label}}).created_at,
                updated_at=Label.from_response({"repository": {"label": label}}).updated_at,
                is_default=label.get("isDefault", False),
                client=self.client
            ) for label in labels_data["nodes"]
        ]
        
        return labels_data
        
    def get_pull_requests(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get pull requests for this repository.
        
        Args:
            first (int, optional): Number of pull requests to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of PullRequest objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set or owner_login/name are not set
            GraphQLError: On GraphQL operation failures
        """
        if not self.client:
            raise ValueError("Client must be set to fetch pull requests")
        
        if not self.owner_login or not self.name:
            raise ValueError("Repository owner_login and name must be set to fetch pull requests")
            
        query = load_query("get_repository_pull_requests")
        variables = {
            "owner": self.owner_login,
            "repo": self.name,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        prs_data = response["repository"]["pullRequests"]
        
        # Convert each node to a PullRequest object
        prs_data["nodes"] = [
            PullRequest.from_response(pr, self.client, self.owner_login, self.name) 
            for pr in prs_data["nodes"]
        ]
        
        return prs_data
    
    def get_issues(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get issues for this repository.
        
        Args:
            first (int, optional): Number of issues to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of Issue objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set or owner_login/name are not set
            GraphQLError: On GraphQL operation failures
        """
        from .issue import Issue
        
        if not self.client:
            raise ValueError("Client must be set to fetch issues")
        
        if not self.owner_login or not self.name:
            raise ValueError("Repository owner_login and name must be set to fetch issues")
            
        query = load_query("get_repository_issues")
        variables = {
            "owner": self.owner_login,
            "repo": self.name,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        issues_data = response["repository"]["issues"]
        
        # Convert each node to an Issue object
        issues_data["nodes"] = [
            Issue.from_response(issue_data, self.client, self.owner_login, self.name) 
            for issue_data in issues_data["nodes"]
        ]
        
        return issues_data
    
    def get_discussions(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get discussions for this repository.
        
        Args:
            first (int, optional): Number of discussions to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of discussion data
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set or owner_login/name are not set
            GraphQLError: On GraphQL operation failures
        """
        if not self.client:
            raise ValueError("Client must be set to fetch discussions")
        
        if not self.owner_login or not self.name:
            raise ValueError("Repository owner_login and name must be set to fetch discussions")
            
        query = load_query("get_repository_discussions")
        variables = {
            "owner": self.owner_login,
            "repo": self.name,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        discussions_data = response["repository"]["discussions"]
        
        return discussions_data
    
    def get_milestones(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get milestones for this repository.
        
        Args:
            first (int, optional): Number of milestones to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of Milestone objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set or owner_login/name are not set
            GraphQLError: On GraphQL operation failures
        """
        from .milestone import Milestone
        
        if not self.client:
            raise ValueError("Client must be set to fetch milestones")
        
        if not self.owner_login or not self.name:
            raise ValueError("Repository owner_login and name must be set to fetch milestones")
            
        query = load_query("get_repository_milestones")
        variables = {
            "owner": self.owner_login,
            "repo": self.name,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        milestones_data = response["repository"]["milestones"]
        
        # Convert each node to a Milestone object
        milestones_data["nodes"] = [
            Milestone(
                id=milestone["id"],
                number=milestone["number"],
                title=milestone["title"],
                state=milestone["state"],
                description=milestone.get("description"),
                due_on=milestone.get("dueOn"),
                created_at=milestone.get("createdAt"),
                closed_at=milestone.get("closedAt"),
                updated_at=milestone.get("updatedAt"),
                url=milestone.get("url"),
                client=self.client
            ) for milestone in milestones_data["nodes"]
        ]
        
        return milestones_data
    
    def get_releases(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get releases for this repository.
        
        Args:
            first (int, optional): Number of releases to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of release data
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set or owner_login/name are not set
            GraphQLError: On GraphQL operation failures
        """
        if not self.client:
            raise ValueError("Client must be set to fetch releases")
        
        if not self.owner_login or not self.name:
            raise ValueError("Repository owner_login and name must be set to fetch releases")
            
        query = load_query("get_repository_releases")
        variables = {
            "owner": self.owner_login,
            "repo": self.name,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        releases_data = response["repository"]["releases"]
        
        return releases_data



"""GitHub GraphQL API client library focused on ProjectsV2 functionality."""

from .client import GraphQLClient
from .projects_v2 import ProjectsV2Client, SearchType
from .exceptions import GraphQLError
from .models.issue import Issue
from .models.view import View  
from .models.user import User
from .models.comment import Comment
from .models.field import Field
from .utils import log_message

__version__ = "0.1.0"
__all__ = [
    "GraphQLClient", "ProjectsV2Client", "GraphQLError",
    "SearchType", 
    "Issue", "View", "User", "Comment", "Field",
    "log_message"
]



"""CrossReferencedEvent model class."""

from typing import Dict, Any
from dataclasses import dataclass
from .event import Event


@dataclass
class CrossReferencedEvent(Event):
    """Event representing an issue being cross-referenced."""

    source_repository_name: str = ""
    source_repository_owner: str = ""
    source_issue_number: int = 0
    source_issue_title: str = ""
    source_url: str = ""

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "CrossReferencedEvent":
        """Create a CrossReferencedEvent instance from API response data."""
        event = super().from_response(data)

        source = data.get("source", {})
        repo = {}
        owner = ""
        if "repository" in source:
            repo = source["repository"] or {}
            if "owner" in repo:
                owner = repo.get("owner", {}).get("login", "")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            source_repository_name=repo.get("name", ""),
            source_repository_owner=owner,
            source_issue_number=source.get("number", 0),
            source_issue_title=source.get("title", ""),
            source_url=source.get("url", ""),
        )



"""LabeledEvent model class."""

from typing import Dict, Any
from dataclasses import dataclass
from .event import Event


@dataclass
class LabeledEvent(Event):
    """Event representing a label being added to an issue."""

    label_name: str = ""
    label_color: str = ""

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "LabeledEvent":
        """Create a LabeledEvent instance from API response data."""
        event = super().from_response(data)

        label = data.get("label", {})

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            label_name=label.get("name", ""),
            label_color=label.get("color", ""),
        )



"""ClosedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class ClosedEvent(Event):
    """Event representing an issue being closed."""

    state_reason: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "ClosedEvent":
        """Create a ClosedEvent instance from API response data."""
        event = super().from_response(data)

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            state_reason=data.get("stateReason"),
        )



"""Base GraphQL client implementation for GitHub's API."""

import requests
from typing import Dict, Any, Optional

from .exceptions import GraphQLError, AuthenticationError, RateLimitError


class GraphQLClient:
    """Base client for making GraphQL requests to GitHub's API."""

    def __init__(self, token: str, api_url: str = "https://api.github.com/graphql"):
        """Initialize the GraphQL client.

        Args:
            token (str): GitHub Personal Access Token
            api_url (str, optional): GitHub GraphQL API endpoint
        """
        self.api_url = api_url
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Authorization": f"Bearer {token}",
                "Accept": "application/vnd.github.v4+json",
            }
        )

    def execute(
        self, query: str, variables: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """Execute a GraphQL query or mutation.

        Args:
            query (str): GraphQL query or mutation string
            variables (dict, optional): Variables for the GraphQL operation

        Returns:
            dict: Response data from the API

        Raises:
            GraphQLError: On general GraphQL errors
            AuthenticationError: On authentication failures
            RateLimitError: When rate limit is exceeded
        """
        payload = {"query": query}
        if variables:
            payload["variables"] = variables

        response = self.session.post(self.api_url, json=payload)

        if response.status_code == 401:
            raise AuthenticationError("Invalid authentication credentials")
        elif response.status_code == 403:
            raise RateLimitError("API rate limit exceeded")

        data = response.json()
        if "errors" in data:
            raise GraphQLError("GraphQL operation failed", data["errors"])

        return data.get("data", {})



"""LockedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class LockedEvent(Event):
    """Event representing an issue being locked."""

    lock_reason: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "LockedEvent":
        """Create a LockedEvent instance from API response data."""
        event = super().from_response(data)

        lock_reason = data.get("lockReason")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            lock_reason=lock_reason,
        )



"""MarkedAsDuplicateEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class MarkedAsDuplicateEvent(Event):
    """Event representing an issue being marked as a duplicate of another issue."""

    canonical_number: Optional[int] = None
    canonical_title: Optional[str] = None
    canonical_url: Optional[str] = None
    canonical_repository_name: Optional[str] = None
    canonical_repository_owner: Optional[str] = None

    duplicate_number: Optional[int] = None
    duplicate_title: Optional[str] = None
    duplicate_url: Optional[str] = None
    duplicate_repository_name: Optional[str] = None
    duplicate_repository_owner: Optional[str] = None
    
    # Support for legacy format
    duplicate_id: Optional[str] = None
    duplicate_name: Optional[str] = None
    duplicate_description: Optional[str] = None
    duplicate_color: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "MarkedAsDuplicateEvent":
        """Create a MarkedAsDuplicateEvent instance from API response data."""
        event = super().from_response(data)

        canonical_number = None
        canonical_title = None
        canonical_url = None
        canonical_repository_name = None
        canonical_repository_owner = None

        duplicate_number = None
        duplicate_title = None
        duplicate_url = None
        duplicate_repository_name = None
        duplicate_repository_owner = None
        
        # Legacy format support
        duplicate_id = None
        duplicate_name = None
        duplicate_description = None
        duplicate_color = None

        if data.get("canonical") and data["canonical"] is not None:
            canonical = data["canonical"]
            canonical_number = canonical.get("number")
            canonical_title = canonical.get("title")
            canonical_url = canonical.get("url")

            if canonical.get("repository"):
                canonical_repository_name = canonical["repository"].get("name")
                if canonical["repository"].get("owner"):
                    canonical_repository_owner = canonical["repository"]["owner"].get(
                        "login"
                    )

        if data.get("duplicate") and data["duplicate"] is not None:
            duplicate = data["duplicate"]
            
            # Check if it's the new format (Issue or PullRequest)
            if "number" in duplicate:
                duplicate_number = duplicate.get("number")
                duplicate_title = duplicate.get("title")
                duplicate_url = duplicate.get("url")
                
                if duplicate.get("repository"):
                    duplicate_repository_name = duplicate["repository"].get("name")
                    if duplicate["repository"].get("owner"):
                        duplicate_repository_owner = duplicate["repository"]["owner"].get(
                            "login"
                        )
            # Legacy format support
            else:
                duplicate_id = duplicate.get("id")
                duplicate_name = duplicate.get("name")
                duplicate_description = duplicate.get("description")
                duplicate_color = duplicate.get("color")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            canonical_number=canonical_number,
            canonical_title=canonical_title,
            canonical_url=canonical_url,
            canonical_repository_name=canonical_repository_name,
            canonical_repository_owner=canonical_repository_owner,
            duplicate_number=duplicate_number,
            duplicate_title=duplicate_title,
            duplicate_url=duplicate_url,
            duplicate_repository_name=duplicate_repository_name,
            duplicate_repository_owner=duplicate_repository_owner,
            duplicate_id=duplicate_id,
            duplicate_name=duplicate_name,
            duplicate_description=duplicate_description,
            duplicate_color=duplicate_color,
        )



"""RenamedTitleEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class RenamedTitleEvent(Event):
    """Event representing an issue's title being renamed."""

    previous_title: Optional[str] = None
    current_title: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "RenamedTitleEvent":
        """Create a RenamedTitleEvent instance from API response data."""
        event = super().from_response(data)

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            previous_title=data.get("previousTitle"),
            current_title=data.get("currentTitle"),
        )



"""Field model classes for GitHub Project fields."""

from typing import Dict, Any, Optional, List
from dataclasses import dataclass
from datetime import datetime

from ..client import GraphQLClient
from ..queries import load_query
from .option import Option


@dataclass
class Field:
    """Represents a field in a GitHub Project (V2)."""

    # Field type constants
    TEXT = "ProjectV2ItemFieldTextValue"
    DATE = "ProjectV2ItemFieldDateValue"
    SINGLE_SELECT = "ProjectV2ItemFieldSingleSelectValue"
    USER = "ProjectV2ItemFieldUserValue"
    REPOSITORY = "ProjectV2ItemFieldRepositoryValue"
    MILESTONE = "ProjectV2ItemFieldMilestoneValue"
    NUMBER = "ProjectV2ItemFieldNumberValue"

    id: str
    name: str
    data_type: str
    type: Optional[str] = None
    value: Optional[Any] = None
    updated: Optional[datetime] = None
    options: Optional[List[Option]] = None
    client: Optional[GraphQLClient] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any], client: Optional[GraphQLClient] = None) -> "Field":
        """Create a Field instance from API response data.

        Args:
            data (Dict[str, Any]): Raw field data from API response
            client (Optional[GraphQLClient]): GraphQL client for API calls

        Returns:
            Field: New Field instance
        """
        options = None

        # If this is a single select field and has options, create Option objects
        if data.get("dataType") == "SINGLE_SELECT" and "options" in data:
            options_data = data.get("options", [])
            options = [Option.from_response(option, client) for option in options_data]

        # Set value and updated if they exist in the data
        value = data.get("value")
        updated = None
        if "updatedAt" in data:
            updated = datetime.fromisoformat(data["updatedAt"].replace("Z", "+00:00"))

        return cls(
            id=data.get("id"),
            name=data.get("name"),
            data_type=data.get("dataType"),
            type=data.get("type"),
            value=value,
            updated=updated,
            options=options,
            client=client
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        id: str
    ) -> "Field":
        """Get a field's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            id (str): Field node ID
            
        Returns:
            Field: Self with populated properties
            
        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_field")
        variables = {
            "id": id
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create a field instance with all properties
        field = Field.from_response(response["node"], client)
        
        # Update this instance's attributes with the fetched data
        self.id = field.id
        self.name = field.name
        self.data_type = field.data_type
        self.type = field.type
        self.value = field.value
        self.updated = field.updated
        self.options = field.options
        self.client = client
        
        return self



"""GraphQL query templates package."""

import os
from jinja2 import Environment, FileSystemLoader, select_autoescape

# Create Jinja2 environment pointing to queries directory
env = Environment(
    loader=FileSystemLoader(os.path.dirname(os.path.abspath(__file__))),
    autoescape=select_autoescape(),
)


def load_query(template_name: str) -> str:
    """Load and return a GraphQL query template.

    Args:
        template_name (str): Name of the file without .graphql extension

    Returns:
        str: The loaded query template
    """
    return env.get_template(f"{template_name}.graphql").render()



"""Team model class for GitHub API responses."""

from dataclasses import dataclass
from typing import Optional, Dict, Any

from ..client import GraphQLClient
from ..queries import load_query


@dataclass(eq=True, frozen=True)
class Team:
    """Represents a GitHub team."""

    id: str
    name: str
    slug: str
    description: Optional[str] = None
    url: str = ""
    privacy: str = ""
    members_count: int = 0
    repositories_count: int = 0
    client: Optional[GraphQLClient] = None

    def __hash__(self):
        """Make Team hashable by its id."""
        return hash(self.id)

    @classmethod
    def from_response(cls, data: Dict[str, Any], client: Optional[GraphQLClient] = None) -> "Team":
        """Create Team instance from GraphQL response.

        Args:
            data (Dict[str, Any]): GraphQL response data containing team info
            client (Optional[GraphQLClient], optional): GraphQL client for API calls. Defaults to None.

        Returns:
            Team: New Team instance
        """
        team_data = data.get("node", data.get("team", {}))
        
        return cls(
            id=team_data.get("id", ""),
            name=team_data.get("name", ""),
            slug=team_data.get("slug", ""),
            description=team_data.get("description"),
            url=team_data.get("url", ""),
            privacy=team_data.get("privacy", ""),
            members_count=team_data.get("membersCount", 0),
            repositories_count=team_data.get("repositoriesCount", 0),
            client=client,
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        org: str,
        slug: str
    ) -> "Team":
        """Get a team's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            org (str): Organization name
            slug (str): Team slug
            
        Returns:
            Team: Self with populated properties
            
        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_team")
        variables = {
            "org": org,
            "slug": slug
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create a team instance with all properties
        team = Team.from_response(response["organization"], client)
        
        # Update this instance's attributes with the fetched data
        object.__setattr__(self, "id", team.id)
        object.__setattr__(self, "name", team.name)
        object.__setattr__(self, "slug", team.slug)
        object.__setattr__(self, "description", team.description)
        object.__setattr__(self, "url", team.url)
        object.__setattr__(self, "privacy", team.privacy)
        object.__setattr__(self, "members_count", team.members_count)
        object.__setattr__(self, "repositories_count", team.repositories_count)
        object.__setattr__(self, "client", client)
        
        return self


"""Option model class for GitHub Project field options."""

from typing import Dict, Any, Optional
from dataclasses import dataclass

from ..client import GraphQLClient
from ..queries import load_query


@dataclass
class Option:
    """Represents an option in a single select field in a GitHub Project (V2)."""

    id: str
    name: str
    client: Optional[GraphQLClient] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any], client: Optional[GraphQLClient] = None) -> "Option":
        """Create an Option instance from API response data.

        Args:
            data (Dict[str, Any]): Raw option data from API response
            client (Optional[GraphQLClient]): GraphQL client for API calls

        Returns:
            Option: New Option instance
        """
        return cls(
            id=data.get("id"),
            name=data.get("name"),
            client=client
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        id: str
    ) -> "Option":
        """Get an option's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            id (str): Option node ID
            
        Returns:
            Option: Self with populated properties
            
        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_option")
        variables = {
            "id": id
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create an option instance with all properties
        option = Option.from_response(response["node"], client)
        
        # Update this instance's attributes with the fetched data
        self.id = option.id
        self.name = option.name
        self.client = client
        
        return self


"""UnassignedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class UnassignedEvent(Event):
    """Event representing a user being unassigned from an issue."""

    assignee_login: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "UnassignedEvent":
        """Create an UnassignedEvent instance from API response data."""
        event = super().from_response(data)

        assignee_login = None
        if data.get("assignee") and data["assignee"] is not None:
            assignee_login = data["assignee"].get("login")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            assignee_login=assignee_login,
        )



"""Project Draft Issue model class."""

from typing import Dict, Any, List, Optional
from dataclasses import dataclass
from datetime import datetime
from ..client import GraphQLClient
from ..queries import load_query
from .field import Field
from .issue import Issue


@dataclass
class DraftIssue(Issue):
    """Represents a draft issue in a GitHub Project (V2)."""

    @classmethod
    def from_response(
        cls,
        data: Dict[str, Any],
        client: Optional[GraphQLClient] = None,
        owner: Optional[str] = None,
        repo: Optional[str] = None,
    ) -> "DraftIssue":
        """Create a DraftIssue instance from API response data.

        Args:
            data (Dict[str, Any]): Raw draft issue data from API response
            client (Optional[GraphQLClient]): GraphQL client for API calls
            owner (Optional[str]): Owner of the repository
            repo (Optional[str]): Repository name

        Returns:
            DraftIssue: New DraftIssue instance
        """
        # Handle DraftIssue which has different fields than regular Issues
        if "content" in data:  # Project item response
            content = data["content"]
            return cls(
                id=content["id"],
                number=0,  # DraftIssue doesn't have a number
                title=content["title"],
                url="",  # DraftIssue doesn't have a URL
                state="",  # DraftIssue doesn't have a state
                created_at=datetime.fromisoformat(
                    content["createdAt"].replace("Z", "+00:00")
                ),
                updated_at=datetime.fromisoformat(
                    content["updatedAt"].replace("Z", "+00:00")
                ),
                author_login=content.get("creator", {}).get("login", ""),
                body=content.get("body"),
                field_values=data.get("fieldValues", {"nodes": []})["nodes"],
                client=client,
                owner=owner,
                repo=repo,
            )
        else:  # Draft issue-specific response
            # For search results the data is in ["node"], directly in data
            if "node" in data:
                issue_data = data["node"]
            else:
                issue_data = data

            return cls(
                id=issue_data["id"],
                number=0,  # DraftIssue doesn't have a number
                title=issue_data["title"],
                url="",  # DraftIssue doesn't have a URL
                state="",  # DraftIssue doesn't have a state
                created_at=datetime.fromisoformat(
                    issue_data["createdAt"].replace("Z", "+00:00")
                ),
                updated_at=datetime.fromisoformat(
                    issue_data["updatedAt"].replace("Z", "+00:00")
                ),
                author_login=issue_data.get("creator", {}).get("login", ""),
                body=issue_data.get("body"),
                field_values=[],  # Not included in issue responses
                client=client,
                owner=owner,
                repo=repo,
            )


"""SubIssueAddedEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class SubIssueAddedEvent(Event):
    """Event representing a sub-issue being added to an issue."""

    child_issue_number: Optional[int] = None
    child_issue_title: Optional[str] = None
    child_issue_url: Optional[str] = None
    child_repository_name: Optional[str] = None
    child_repository_owner: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "SubIssueAddedEvent":
        """Create a SubIssueAddedEvent instance from API response data."""
        event = super().from_response(data)

        child_issue_number = None
        child_issue_title = None
        child_issue_url = None
        child_repository_name = None
        child_repository_owner = None

        if data.get("subIssue") and data["subIssue"].get("__typename") == "Issue":
            child_issue = data["subIssue"]
            child_issue_number = child_issue.get("number")
            child_issue_title = child_issue.get("title")
            child_issue_url = child_issue.get("url")

            if child_issue.get("repository"):
                child_repository_name = child_issue["repository"].get("name")
                if child_issue["repository"].get("owner"):
                    child_repository_owner = child_issue["repository"]["owner"].get(
                        "login"
                    )

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            child_issue_number=child_issue_number,
            child_issue_title=child_issue_title,
            child_issue_url=child_issue_url,
            child_repository_name=child_repository_name,
            child_repository_owner=child_repository_owner,
        )



"""GitHub API clients for different types of data."""

from typing import Dict, Any
from enum import Enum
from .client import GraphQLClient


class SearchType(str, Enum):
    """Supported GitHub search types."""
    ISSUE = "ISSUE"
    REPOSITORY = "REPOSITORY"


def _paginate_query(
    client: GraphQLClient,
    query: str,
    variables: Dict[str, Any],
    data_path: str,
    nested_path: str = None,
) -> Dict[str, Any]:
    """Helper method to handle pagination in GraphQL queries.

    Args:
        client (GraphQLClient): GraphQL client
        query (str): GraphQL query string
        variables (Dict[str, Any]): Query variables
        data_path (str): Path to nodes array in response (e.g. "node.fields" or "node.items")
        nested_path: Path to nested nodes for nested pagination (e.g. "fieldValues")

    Returns:
        Dict[str, Any]: Merged results from all pages
    """
    all_nodes = []
    has_next_page = True
    after = None

    while has_next_page:
        curr_vars = {**variables}
        if after:
            curr_vars["after"] = after

        result = client.execute(query, curr_vars)

        # Navigate to the nodes data using data_path
        curr_data = result
        for key in data_path.split("."):
            curr_data = curr_data.get(key, {})

        nodes = curr_data.get("nodes", [])

        # Handle nested pagination if requested
        if nested_path:
            for node in nodes:
                curr_nested_data = node.get(nested_path, {})
                has_next_nested = curr_nested_data.get("pageInfo", {}).get(
                    "hasNextPage", False
                )
                nested_cursor = curr_nested_data.get("pageInfo", {}).get(
                    "endCursor"
                )
                nested_nodes = curr_nested_data.get("nodes", [])

                all_nested_nodes = nested_nodes.copy()

                # Paginate through nested nodes
                while has_next_nested:
                    curr_vars = {**variables}  # Start fresh
                    curr_vars.update({"fieldsAfter": nested_cursor})
                    nested_result = client.execute(query, curr_vars)

                    # Navigate to same location
                    curr_nested_result = nested_result
                    for key in data_path.split("."):
                        curr_nested_result = curr_nested_result.get(key, {})

                    # Get first node's fieldValues since we're paginating its fields
                    curr_nested_data = curr_nested_result["nodes"][0].get(
                        nested_path, {}
                    )

                    nested_nodes = curr_nested_data.get("nodes", [])
                    all_nested_nodes.extend(nested_nodes)

                    has_next_nested = curr_nested_data.get("pageInfo", {}).get(
                        "hasNextPage", False
                    )
                    nested_cursor = curr_nested_data.get("pageInfo", {}).get(
                        "endCursor"
                    )

                # Update node with merged nested nodes
                node[nested_path]["nodes"] = all_nested_nodes

        all_nodes.extend(nodes)

        page_info = curr_data.get("pageInfo", {})
        has_next_page = page_info.get("hasNextPage", False)
        after = page_info.get("endCursor")

    # Reconstruct response with all nodes
    response = result  # Start with last response
    curr_data = response
    *parent_keys, last_key = data_path.split(".")

    # Navigate to the parent object to update
    for key in parent_keys:
        curr_data = curr_data[key]

    curr_data[last_key]["nodes"] = all_nodes
    return response


"""TransferredEvent model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from .event import Event


@dataclass
class TransferredEvent(Event):
    """Event representing an issue being transferred to another repository."""

    from_repository_name: Optional[str] = None
    from_repository_owner: Optional[str] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "TransferredEvent":
        """Create a TransferredEvent instance from API response data."""
        event = super().from_response(data)

        from_repository_name = None
        from_repository_owner = None

        if data.get("fromRepository"):
            from_repository_name = data["fromRepository"].get("name")
            if data["fromRepository"].get("owner"):
                from_repository_owner = data["fromRepository"]["owner"].get("login")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            from_repository_name=from_repository_name,
            from_repository_owner=from_repository_owner,
        )



"""User model class."""

from typing import Dict, Any, Optional
from dataclasses import dataclass
from datetime import datetime

from ..client import GraphQLClient
from ..queries import load_query


@dataclass
class User:
    """Represents a GitHub User with basic profile information."""

    id: str
    login: str
    name: Optional[str]
    email: Optional[str]
    bio: Optional[str]
    company: Optional[str]
    location: Optional[str]
    website_url: Optional[str]
    url: str
    avatar_url: str
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    client: Optional[GraphQLClient] = None

    def __post_init__(self):
        """Initialize default values for datetime fields."""
        if self.created_at is None:
            self.created_at = datetime.now()
        if self.updated_at is None:
            self.updated_at = datetime.now()

    @classmethod
    def from_response(cls, data: Dict[str, Any], client: Optional[GraphQLClient] = None) -> "User":
        """Create a User instance from API response data.

        Args:
            data (Dict[str, Any]): Raw user data from API response
            client (Optional[GraphQLClient]): GraphQL client for API calls

        Returns:
            User: New User instance
        """
        # Handle responses from node query or direct user query
        if "node" in data:
            user_data = data["node"]
        else:
            # For responses that have user data nested under 'user' key
            user_data = data.get("user", data)

        return cls(
            id=user_data["id"],
            login=user_data["login"],
            name=user_data.get("name"),
            email=user_data.get("email"),
            bio=user_data.get("bio"),
            company=user_data.get("company"),
            location=user_data.get("location"),
            website_url=user_data.get("websiteUrl"),
            url=user_data["url"],
            avatar_url=user_data["avatarUrl"],
            created_at=datetime.fromisoformat(
                user_data["createdAt"].replace("Z", "+00:00")
            ),
            updated_at=datetime.fromisoformat(
                user_data["updatedAt"].replace("Z", "+00:00")
            ),
            client=client,
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        id: Optional[str] = None,
        username: Optional[str] = None
    ) -> "User":
        """Get a user's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            id (Optional[str]): User ID (for lookup by ID)
            username (Optional[str]): Username (for lookup by username)
            
        Returns:
            User: Self with populated properties
            
        Raises:
            ValueError: If neither id nor username is provided
            GraphQLError: On GraphQL operation failures
        """
        if not id and not username:
            raise ValueError("Either id or username must be provided")
            
        if id:
            query = load_query("get_user_by_id")
            variables = {"id": id}
        else:
            query = load_query("get_user")
            variables = {"login": username}
        
        response = client.execute(query, variables)
        
        # Use from_response to create a user instance with all properties
        user = User.from_response(response, client)
        
        # Update this instance's attributes with the fetched data
        self.id = user.id
        self.login = user.login
        self.name = user.name
        self.email = user.email
        self.bio = user.bio
        self.company = user.company
        self.location = user.location
        self.website_url = user.website_url
        self.url = user.url
        self.avatar_url = user.avatar_url
        self.created_at = user.created_at
        self.updated_at = user.updated_at
        self.client = client
        
        return self
        
    def get_followers(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get followers of this user.
        
        Args:
            first (int, optional): Number of followers to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of User objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set
            GraphQLError: On GraphQL operation failures
        """
        if not self.client:
            raise ValueError("Client must be set to fetch followers")
            
        query = load_query("get_user_followers")
        variables = {
            "login": self.login,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        followers_data = response["user"]["followers"]
        
        # Convert each node to a User object
        followers_data["nodes"] = [
            User.from_response({"user": follower}, self.client) 
            for follower in followers_data["nodes"]
        ]
        
        return followers_data
        
    def get_following(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get users followed by this user.
        
        Args:
            first (int, optional): Number of following users to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of User objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set
            GraphQLError: On GraphQL operation failures
        """
        if not self.client:
            raise ValueError("Client must be set to fetch following users")
            
        query = load_query("get_user_following")
        variables = {
            "login": self.login,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        following_data = response["user"]["following"]
        
        # Convert each node to a User object
        following_data["nodes"] = [
            User.from_response({"user": following}, self.client) 
            for following in following_data["nodes"]
        ]
        
        return following_data
        
    def get_issues(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get issues assigned to this user.
        
        Args:
            first (int, optional): Number of issues to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of Issue objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set
            GraphQLError: On GraphQL operation failures
        """
        from .issue import Issue
        
        if not self.client:
            raise ValueError("Client must be set to fetch user issues")
            
        query = load_query("get_user_issues")
        variables = {
            "login": self.login,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        issues_data = response["user"]["issues"]
        
        # Convert each node to an Issue object
        issues_data["nodes"] = [
            Issue(
                id=issue_data["id"],
                number=issue_data["number"],
                title=issue_data["title"],
                url=issue_data["url"],
                state=issue_data["state"],
                created_at=datetime.fromisoformat(issue_data["createdAt"].replace("Z", "+00:00")),
                updated_at=datetime.fromisoformat(issue_data["updatedAt"].replace("Z", "+00:00")),
                author_login=issue_data["author"]["login"],
                body=issue_data.get("body"),
                field_values=[],
                client=self.client,
                owner=issue_data["repository"]["owner"]["login"],
                repo=issue_data["repository"]["name"]
            )
            for issue_data in issues_data["nodes"]
        ]
        
        return issues_data



"""Project model class."""

from typing import Dict, Any, Optional, List, Union, TYPE_CHECKING
from dataclasses import dataclass
from datetime import datetime
from ..client import GraphQLClient
from ..clients import _paginate_query
from ..queries import load_query
from .issue import Issue
from .view import View
from .repository import Repository
from .team import Team
from .workflow import Workflow
from .user import User
from .organization import Organization

if TYPE_CHECKING:
    from .field import Field
    from .draft_issue import DraftIssue
    from .pull_request import PullRequest


@dataclass
class Project:
    """Represents a GitHub Project (V2) with its properties and methods."""

    id: str = ""
    title: str = ""
    short_description: Optional[str] = None
    public: bool = False
    closed: bool = False
    url: str = ""
    number: int = 0
    creator: Optional[Union[User, Organization]] = None
    owner: Optional[Union[User, Organization]] = None
    readme: Optional[str] = None
    updated_at: Optional[datetime] = None
    resource_path: Optional[str] = None
    client: Optional[GraphQLClient] = None

    @classmethod
    def from_response(cls, data: Dict[str, Any], client: GraphQLClient) -> "Project":
        """Create a Project instance from API response data.

        Args:
            data (Dict[str, Any]): Raw project data from API response
            client (GraphQLClient): GraphQL client for API calls

        Returns:
            Project: New Project instance
        """
        from .utils import parse_datetime

        # Get the project data - could be under user or organization key
        owner_data = data.get("user") or data.get("organization", {})
        project_data = owner_data.get("projectV2", {})

        # Parse creator field
        creator = None
        if project_data.get("creator"):
            creator_data = project_data["creator"]
            if (
                creator_data.get("__typename") == "Organization"
                or "organization" in creator_data
            ):
                creator = Organization.from_response(
                    {"organization": creator_data}, client
                )
            else:
                creator = User.from_response({"user": creator_data}, client)

        # Parse owner field
        owner = None
        if project_data.get("owner"):
            owner_data = project_data["owner"]
            print(owner_data)
            
            if (
                owner_data.get("__typename") == "Organization"
                or "organization" in owner_data
            ):
                owner = Organization.from_response({"organization": owner_data}, client)
            else:
                owner = User.from_response({"user": owner_data}, client)

        return cls(
            id=project_data.get("id"),
            title=project_data.get("title"),
            short_description=project_data.get("shortDescription"),
            public=project_data.get("public", False),
            closed=project_data.get("closed", False),
            url=project_data.get("url"),
            number=project_data.get("number"),
            creator=creator,
            owner=owner,
            readme=project_data.get("readme"),
            updated_at=parse_datetime(project_data.get("updatedAt")),
            resource_path=project_data.get("resourcePath"),
            client=client,
        )

    def get(self, *, client: GraphQLClient, owner: str, number: int) -> "Project":
        """Get a project's details and populate this instance.

        Args:
            client (GraphQLClient): GraphQL client for API calls
            owner (str): Project owner (organization or username)
            number (int): Project number

        Returns:
            Project: Self with populated properties

        Raises:
            GraphQLError: On GraphQL operation failures
        """
        from gh_project_v2.projects_v2 import ProjectsV2Client

        projects_client = ProjectsV2Client(client)
        project = projects_client.get_project(number, owner)

        # Update this instance's attributes with the fetched data
        self.id = project.id
        self.title = project.title
        self.short_description = project.short_description
        self.public = project.public
        self.closed = project.closed
        self.url = project.url
        self.number = project.number
        self.creator = project.creator
        self.owner = project.owner
        self.readme = project.readme
        self.updated_at = project.updated_at
        self.resource_path = project.resource_path
        self.client = client

        return self

    # Removed duplicate _paginate_query method - now using the one from clients.py

    def get_fields(self, first: int = 20) -> List["Field"]:
        """Get fields configured in this project.

        This method is deprecated. Use ProjectsV2Client.list_project_fields instead.

        Args:
            first (int, optional): Number of items per page. Defaults to 20.

        Returns:
            List[Field]: List of project field objects
        """
        from gh_project_v2.projects_v2 import ProjectsV2Client

        projects_client = ProjectsV2Client(self.client)
        return projects_client.list_project_fields(self.id, first)

    def find_field_by_name(
        self, name: str, fields: Optional[List["Field"]] = None
    ) -> Optional["Field"]:
        """Search for a field by name in this project.

        Args:
            name (str): The name of the field to search for
            fields (Optional[List["Field"]], optional): List of fields to search in.
                If None, will fetch fields from the project. Defaults to None.

        Returns:
            Optional["Field"]: Matching field or None if not found
        """
        if fields is None:
            from gh_project_v2.projects_v2 import ProjectsV2Client

            projects_client = ProjectsV2Client(self.client)
            fields = projects_client.list_project_fields(self.id)

        for field in fields:
            if field.name == name:
                return field

        return None

    def get_items(
        self, first: int = 20, fields_first: Optional[int] = None
    ) -> List[Union["Issue", "DraftIssue", "PullRequest"]]:
        """Get items (issues/PRs/draft issues) in this project.

        Args:
            first (int, optional): Number of items per page. Defaults to 20.
            fields_first (int, optional): Number of field values per page. Default None.
                A value between 1 and 100 must be provided if set.

        Returns:
            List[Union[Issue, DraftIssue, PullRequest]]: List of project item objects
                from the API

        Raises:
            ValueError: If fields_first is greater than 100 or less than 1
        """
        if fields_first is not None:
            if not 1 <= fields_first <= 100:
                raise ValueError("fields_first must be between 1 and 100")

        query = load_query("get_project_issues")
        variables = {"projectId": self.id, "first": first}
        if fields_first is not None:
            variables["fieldsFirst"] = fields_first
            response = _paginate_query(
                self.client, query, variables, "node.items", nested_path="fieldValues"
            )
        else:
            response = _paginate_query(self.client, query, variables, "node.items")
        items_data = response["node"]["items"]

        from .draft_issue import DraftIssue
        from .issue import Issue
        from .pull_request import PullRequest

        items = []
        for node in items_data["nodes"]:
            item_type = node["content"].get("__typename")
            if item_type == "PullRequest":
                items.append(PullRequest.from_response(node))
            elif item_type == "DraftIssue":
                items.append(DraftIssue.from_response(node))
            else:  # Default to Issue for backward compatibility
                items.append(Issue.from_response(node))

        return items

    def get_views(
        self, first: int = 20, fields_first: Optional[int] = None
    ) -> List[View]:
        """Get views configured in this project.

        Args:
            first (int, optional): Number of views to fetch per page. Defaults to 20.
            fields_first (int, optional): Number of fields per page. Defaults to 20.
                A value between 1 and 100 must be provided if set.

        Returns:
            List[View]: List of project views as View objects

        Raises:
            ValueError: If fields_first is greater than 100 or less than 1
        """
        if fields_first is not None:
            if not 1 <= fields_first <= 100:
                raise ValueError("fields_first must be between 1 and 100")

        query = load_query("get_project_views")
        variables = {"projectId": self.id, "first": first}
        if fields_first is not None:
            variables["fieldsFirst"] = fields_first
            response = _paginate_query(
                self.client, query, variables, "node.views", nested_path="fields"
            )
        else:
            response = _paginate_query(self.client, query, variables, "node.views")
        raw_views = response["node"]["views"]
        return [View.from_response(node) for node in raw_views["nodes"]]

    def get_repositories(self, first: int = 20) -> List[Repository]:
        """Get repositories linked to this project.

        Args:
            first (int, optional): Number of repositories to fetch per page.
                Defaults to 20.

        Returns:
            List[Repository]: List of Repository instances with data from the API
        """
        query = load_query("get_project_repositories")
        variables = {"projectId": self.id, "first": first}
        response = _paginate_query(self.client, query, variables, "node.repositories")

        repositories_data = response["node"]["repositories"]
        return [
            Repository.from_response({"repository": node})
            for node in repositories_data["nodes"]
        ]

    def get_teams(self, first: int = 20) -> List[Team]:
        """Get teams linked to this project.

        Args:
            first (int, optional): Number of teams to fetch per page. Defaults to 20.

        Returns:
            List[Team]: List of Team instances with data from the API
        """
        query = load_query("get_project_teams")
        variables = {"projectId": self.id, "first": first}
        response = _paginate_query(self.client, query, variables, "node.teams")

        teams_data = response["node"]["teams"]
        return [
            Team.from_response({"team": node}, self.client)
            for node in teams_data["nodes"]
        ]

    def get_workflows(self, first: int = 20) -> List[Workflow]:
        """Get workflows linked to this project.

        Args:
            first (int, optional): Number of workflows to fetch per page.
                Defaults to 20.

        Returns:
            List[Workflow]: List of Workflow instances with data from the API
        """
        query = load_query("get_project_workflows")
        variables = {"projectId": self.id, "first": first}
        response = _paginate_query(self.client, query, variables, "node.workflows")

        workflows_data = response["node"]["workflows"]
        return [
            Workflow.from_response({"workflow": node}, self.client)
            for node in workflows_data["nodes"]
        ]



"""AssignedEvent model class."""

from typing import Dict, Any
from dataclasses import dataclass
from .event import Event


@dataclass
class AssignedEvent(Event):
    """Event representing an issue being assigned to a user."""

    assignee_login: str = ""

    @classmethod
    def from_response(cls, data: Dict[str, Any]) -> "AssignedEvent":
        """Create an AssignedEvent instance from API response data."""
        event = super().from_response(data)

        assignee_login = ""
        if data.get("assignee") and data["assignee"] is not None:
            assignee_login = data["assignee"].get("login", "")

        return cls(
            id=event.id,
            created_at=event.created_at,
            actor_login=event.actor_login,
            assignee_login=assignee_login,
        )



"""Models representing GitHub Milestones."""

from dataclasses import dataclass
from typing import Optional, Dict, Any
from datetime import datetime

from ..client import GraphQLClient
from ..queries import load_query
from .utils import parse_datetime

@dataclass
class Milestone:
    """Represents a GitHub milestone."""

    id: str
    number: int
    title: str
    state: str
    description: Optional[str] = None
    due_on: Optional[datetime] = None
    created_at: Optional[datetime] = None
    closed_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    url: Optional[str] = None
    client: Optional[GraphQLClient] = None

    """Initialize a new Milestone.

    Args:
        id (str): The GitHub node ID
        number (int): The milestone number
        title (str): The milestone title
        state (str): Current state (OPEN or CLOSED)
        description (Optional[str]): Description text
        due_on (Optional[datetime]): Due date if set
        created_at (Optional[datetime]): Creation timestamp
        closed_at (Optional[datetime]): Close timestamp if closed
        updated_at (Optional[datetime]): Last update timestamp
        url (Optional[str]): URL to view the milestone
        client (Optional[GraphQLClient]): GraphQL client for API calls
    """

    @classmethod
    def from_response(cls, response: dict, client: Optional[GraphQLClient] = None) -> Optional["Milestone"]:
        """Create a Milestone from a GraphQL response.

        Args:
            response (dict): The GraphQL response data
            client (Optional[GraphQLClient]): GraphQL client for API calls

        Returns:
            Optional[Milestone]: A Milestone object or None if not found
        """
        repo_data = response.get("repository", {})
        milestone_data = repo_data.get("milestone")

        if not milestone_data:
            return None

        return cls(
            id=milestone_data["id"],
            number=milestone_data["number"],
            title=milestone_data["title"],
            state=milestone_data["state"],
            description=milestone_data.get("description"),
            due_on=parse_datetime(milestone_data.get("dueOn")),
            created_at=parse_datetime(milestone_data.get("createdAt")),
            closed_at=parse_datetime(milestone_data.get("closedAt")),
            updated_at=parse_datetime(milestone_data.get("updatedAt")),
            url=milestone_data.get("url"),
            client=client,
        )
        
    def get(
        self,
        *,
        client: GraphQLClient,
        owner: str,
        repo: str,
        number: int
    ) -> Optional["Milestone"]:
        """Get a milestone's details and populate this instance.
        
        Args:
            client (GraphQLClient): GraphQL client for API calls
            owner (str): Repository owner (organization or username)
            repo (str): Repository name
            number (int): Milestone number
            
        Returns:
            Optional[Milestone]: Self with populated properties or None if not found
            
        Raises:
            GraphQLError: On GraphQL operation failures
        """
        query = load_query("get_milestone")
        variables = {
            "owner": owner,
            "repo": repo,
            "number": number
        }
        
        response = client.execute(query, variables)
        
        # Use from_response to create a milestone instance with all properties
        milestone = Milestone.from_response(response, client)
        
        if milestone is None:
            return None
            
        # Update this instance's attributes with the fetched data
        self.id = milestone.id
        self.number = milestone.number
        self.title = milestone.title
        self.state = milestone.state
        self.description = milestone.description
        self.due_on = milestone.due_on
        self.created_at = milestone.created_at
        self.closed_at = milestone.closed_at
        self.updated_at = milestone.updated_at
        self.url = milestone.url
        self.client = client
        
        return self
        
    def get_issues(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get issues for this milestone.
        
        Args:
            first (int, optional): Number of issues to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of Issue objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set
            GraphQLError: On GraphQL operation failures
        """
        from .issue import Issue
        
        if not self.client:
            raise ValueError("Client must be set to fetch milestone issues")
            
        # Extract owner and repo from URL
        # URL format: https://github.com/owner/repo/milestone/number
        url_parts = self.url.split('/')
        if len(url_parts) < 5 or url_parts[2] != 'github.com':
            raise ValueError("Invalid milestone URL format")
            
        owner = url_parts[3]
        repo = url_parts[4]
            
        query = load_query("get_milestone_issues")
        variables = {
            "owner": owner,
            "repo": repo,
            "number": self.number,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        issues_data = response["repository"]["milestone"]["issues"]
        
        # Convert each node to an Issue object
        issues_data["nodes"] = [
            Issue.from_response(issue_data, self.client, owner, repo) 
            for issue_data in issues_data["nodes"]
        ]
        
        return issues_data
        
    def get_pull_requests(self, first: int = 20, after: Optional[str] = None) -> Dict[str, Any]:
        """Get pull requests for this milestone.
        
        Args:
            first (int, optional): Number of pull requests to fetch per page. Defaults to 20.
            after (Optional[str], optional): Cursor for pagination. Defaults to None.
            
        Returns:
            Dict[str, Any]: Dictionary containing:
                - "nodes": List of PullRequest objects
                - "pageInfo": PageInfo object with hasNextPage and endCursor
                
        Raises:
            ValueError: If client is not set
            GraphQLError: On GraphQL operation failures
        """
        from .pull_request import PullRequest
        
        if not self.client:
            raise ValueError("Client must be set to fetch milestone pull requests")
            
        # Extract owner and repo from URL
        # URL format: https://github.com/owner/repo/milestone/number
        url_parts = self.url.split('/')
        if len(url_parts) < 5 or url_parts[2] != 'github.com':
            raise ValueError("Invalid milestone URL format")
            
        owner = url_parts[3]
        repo = url_parts[4]
            
        query = load_query("get_milestone_pull_requests")
        variables = {
            "owner": owner,
            "repo": repo,
            "number": self.number,
            "first": first
        }
        
        if after:
            variables["after"] = after
            
        response = self.client.execute(query, variables)
        prs_data = response["repository"]["milestone"]["pullRequests"]
        
        # Convert each node to a PullRequest object
        prs_data["nodes"] = [
            PullRequest.from_response(pr, self.client, owner, repo) 
            for pr in prs_data["nodes"]
        ]
        
        return prs_data
        
    def get_repo(self) -> Optional["Repository"]:
        """Get the repository associated with this milestone.
        
        Returns:
            Optional[Repository]: Repository object or None if not found
            
        Raises:
            ValueError: If client is not set
            GraphQLError: On GraphQL operation failures
        """
        from .repository import Repository
        
        if not self.client:
            raise ValueError("Client must be set to fetch milestone repository")
            
        # Extract owner and repo from URL
        # URL format: https://github.com/owner/repo/milestone/number
        url_parts = self.url.split('/')
        if len(url_parts) < 5 or url_parts[2] != 'github.com':
            raise ValueError("Invalid milestone URL format")
            
        owner = url_parts[3]
        repo = url_parts[4]
            
        query = load_query("get_milestone_repository")
        variables = {
            "owner": owner,
            "repo": repo,
            "number": self.number
        }
        
        response = self.client.execute(query, variables)
        
        milestone_data = response.get("repository", {}).get("milestone")
        if not milestone_data:
            return None
            
        repo_data = milestone_data.get("repository")
        if not repo_data:
            return None
            
        owner_data = repo_data.get("owner", {})
            
        return Repository(
            id=repo_data["id"],
            name=repo_data["name"],
            name_with_owner=repo_data["nameWithOwner"],
            url=repo_data["url"],
            owner_login=owner_data["login"],
            owner_url=owner_data["url"],
            description=repo_data.get("description"),
            is_private=repo_data.get("isPrivate", False),
            homepage_url=repo_data.get("homepageUrl"),
            client=self.client
        )